#include <qmessagebox.h>
#include <qstringlist.h>
#include <cstdlib>

#include "activeplaylist.h"
#include "jackctrl.h"

extern global_info gi;

ActivePlaylist::ActivePlaylist(QWidget *p, QScrollBar *scb) : QWidget(p),
	aScroll(scb),
	spacer(new QLabel(this)),
	timerFast(new QTimer(this)), 
	timerSlow(new QTimer(this)), 
	parent(p), isUpdatingGraphics(false), 
	itemFirst(0), itemLast(0), itemCount(0) {

	aScroll->setEnabled(false);
	aScroll->setMinValue(0);
	aScroll->setValue(0);

	connect(aScroll, SIGNAL(valueChanged(int)), this, SLOT(updateLayout()));
	connect(timerFast, SIGNAL(timeout()), this, SLOT(updateData()));
	connect(timerSlow, SIGNAL(timeout()), this, SLOT(updateGraphics()));

	timerFast->start(TIMER_FAST_DELAY);
	timerSlow->start(TIMER_SLOW_DELAY);
}

ActivePlaylist::~ActivePlaylist() {
	delete spacer;
	delete timerFast;
	delete timerSlow;
}

void ActivePlaylist::updateScroll() {
	if(!isVisible())
		return;

	int height = 0;
	uint num = 0;

	for(ActiveEntry *it = itemFirst; it; it = it->next) {
		height += it->sizeHint().height();
		if(height >= parent->height())
			break;
		num++;
	}

	if(num < itemCount) {
		aScroll->setMaxValue(itemCount - num);
		aScroll->setEnabled(true);
	}
	else {
		aScroll->setMaxValue(0);
		aScroll->setEnabled(false);
	}

	updateLayout();
}

void ActivePlaylist::updateLayout() {
	if(!isVisible())
		return;

	int offset = 0;
	ActiveEntry *it;

	if(aScroll->isEnabled()) {
		int i = 0;
		int skip = aScroll->value();

		it = itemFirst;
		while(i < skip && it != 0) {
			it->hide();
			it = it->next;
			i++;
		}

		while(offset < parent->height() && it != 0) {
			it->move(0, offset);
			it->show();
			offset += it->height();
			it = it->next;
		}

		
		while(it) {
			it->hide();
			it = it->next;
		}
	}
	else {
		for(it = itemFirst; it; it = it->next) {
			it->move(0, offset);
			it->show();
			offset += it->height();
		}
	}
}


void ActivePlaylist::updateGraphics() {
	if(!isVisible() || itemCount == 0 || isUpdatingGraphics)
		return;
	isUpdatingGraphics = true;

	ActiveEntry *it = itemFirst;
	while(it) {
		it->updateWidgets();
		it = it->next;
	}

	isUpdatingGraphics = false;
}

void ActivePlaylist::updateData() {
	for(ActiveEntry *it = itemFirst; it; it = it->next)
		it->hasBeenUpdated = false;

	if(gi.client) {
		if(get_rdlock(&gi)) {
			audio_entry *entry = gi.audioList;
			while(entry) {
				if(entry->gui_ptr)
					((ActiveEntry *)  (entry->gui_ptr))->updateData();

				entry = entry->next;
			}
			release_rdlock(&gi);
		}
	}

	ActiveEntry *del;
	ActiveEntry *it = itemFirst;
	while(it) {
		if(it->hasBeenUpdated)
			it = it->next;
		else {
			del = it;
			it = it->next;
			remEntry(del);
		}
	}
}

void ActivePlaylist::resizeEvent(QResizeEvent *) {
	updateScroll();
}

void ActivePlaylist::addEntry(ActiveEntry *entry) {
	ActiveEntry *it;

	if(itemCount == 0) {
		itemFirst = entry;
		itemLast = entry;
		entry->next = 0;
		entry->prev = 0;
	}
	else if(entry->index == 0) {
		entry->next = itemFirst;
		if(itemFirst)
			itemFirst->prev = entry;
		itemFirst = entry;
		entry->prev = 0;
	}
	else if(entry->index < itemCount) {
		for(it = itemFirst; it; it = it->next)
			if(it->index == entry->index)
				break;

		entry->next = it;
		entry->prev = it->prev;
		entry->next->prev = entry;
		entry->prev->next = entry;
	}
	else {
		entry->prev = itemLast;
		if(itemLast)
			itemLast->next = entry;
		itemLast = entry;
		entry->next = 0;
	}
	itemCount++;		

	uint i = 0;
	for(it = itemFirst; it; it = it->next)
		it->index = i++;

	updateGraphics();
	updateScroll();
}

void ActivePlaylist::stop(ActiveEntry *entry) {
	take_out_audio_entry(entry->audioEntry);
}

void ActivePlaylist::stop(QString name) {
	ActiveEntry *it = itemFirst, *del;

	while(it) {
		if(it->label_name->text() == name) {
			del = it;
			it = it->next;
			stop(del);
		}
		else
			it = it->next;
	}

	updateGraphics();
	updateScroll();
}

void ActivePlaylist::killAll() {
	ActiveEntry *n;
	for(ActiveEntry *it = itemFirst; it; it = n) {
		take_out_audio_entry(it->audioEntry);
		it->hide();
		n = it->next;
		it->vuMeter->timer->stop();
		delete it;
	}

	itemFirst = 0;
	itemLast = 0;
	itemCount = 0;

	updateGraphics();
	updateScroll();
}

void ActivePlaylist::killNum(int num) {
	ActiveEntry *it = itemFirst;
	int n = 0;
	while(n < num && it != 0) {
		it = it->next;
		n++;
	}

	if(it)
		stop(it);	
}

void ActivePlaylist::remEntry(ActiveEntry *entry) {
	entry->hide();
	entry->vuMeter->timer->stop();
	remFromList(entry);
	delete entry;

	updateGraphics();
	updateScroll();
}

void ActivePlaylist::remFromList(ActiveEntry *entry) {
	if(itemCount == 1) {
		itemFirst = 0;
		itemLast = 0;
	}
	else if(entry == itemFirst) {
		itemFirst = entry->next;
		itemFirst->prev = 0;
	}
	else if(entry == itemLast) {
		itemLast = entry->prev;
		itemLast->next = 0;
	}
	else {
		entry->prev->next = entry->next;
		entry->next->prev = entry->prev;
	}
	itemCount--;

	uint i = 0;
	for(ActiveEntry *it = itemFirst; it; it = it->next)
		it->index = i++;
}

void ActivePlaylist::openEntry(QString filename, QString name, ChannelPortList chnVSport, int volume, signed int repeat_count) {
	if(filename.isEmpty())
		return;

	port_connection *pc = 0, *ptrPc;
	ChannelPortList::iterator it;

	if(chnVSport.count()) {
		pc = (port_connection *) malloc(sizeof(port_connection) * chnVSport.count());
		ptrPc = pc;

		it = chnVSport.begin();
		while(it != chnVSport.end()) {
			ptrPc->last = 0;
			ptrPc->srcChannel = (*it).channel;
			ptrPc->dstName = (char *) (*it).name.ascii();
			it++;
			ptrPc++;
		 }

		ptrPc--;
		ptrPc->last = 1;
	}

	audio_entry *ae = open_audio_entry(&gi, filename.ascii(), name.ascii(), pc, (float) volume / 100.0, repeat_count);

	free(pc);

	if(ae) {
		if(ae->error) {
			QStringList msg;
			if(ae->error & ERR_NO_JACK_CLIENT)
				msg << "No registered as a jack client!";

			if(ae->error & ERR_COULD_NOT_ALLOCATE_MEM)
				msg << "Could not allocate enough memory!";

			if(ae->error & ERR_TO_LONG_CLIENT_NAME)
				msg << "Client name too long!";

			if(ae->error & ERR_MISSING_FILENAME)
				msg << "Missing filename!";

			if(ae->error & ERR_FILE_NOT_FOUND)
				msg << "File not found!";

			if(ae->error & ERR_TO_MANY_CHANNELS)
				msg << "Too many channels in audio file!";

			if(ae->error & ERR_SETTING_NORMALIZE)
				msg << "Could not set the SF_NORMALIZE flag!";

			if(ae->error & ERR_REG_JACK_PORT)
				msg << "Could not register jack ports!";

			if(ae->error & ERR_GET_PORT_INFO)
				msg << "Error while requsting jack for port info!";

			if(ae->error & ERR_COULD_NOT_RINGBUF)
				msg << "Could not allocate ringbuffers!";

			if(ae->error & ERR_FILE_COULD_NOT_SFOPEN)
				msg << "Could not open file!";

			if(msg.count() == 0)
				msg << "An unknown error happened!";

			destroy_audio_entry(ae);

			QMessageBox::critical(this, "Error!", msg.join("\n"), QMessageBox::Ok, QMessageBox::NoButton);
		}
		else {
			ActiveEntry *n = new ActiveEntry(this, ae, name, chnVSport, volume);

			ae->gui_ptr = n;
			addEntry(n);
		}
	}
	else
		QMessageBox::critical(this, "Error!", "Could not allocate enough memory!", QMessageBox::Ok, QMessageBox::NoButton);
}

void ActivePlaylist::timersStop() {
	timerSlow->stop();
	timerFast->stop();

	for(ActiveEntry *it = itemFirst; it; it = it->next)
		it->vuMeter->timer->stop();
}

void ActivePlaylist::timersStart() {
	timerFast->start(TIMER_FAST_DELAY);
	timerSlow->start(TIMER_SLOW_DELAY);

	for(ActiveEntry *it = itemFirst; it; it = it->next)
		 it->vuMeter->timer->start(TIMER_VU_DELAY);
}
