#include <qfont.h>
#include <qmessagebox.h>
#include <qcolordialog.h>

#include "misc.h"
#include "configwidgets.h"
#include "config.h"
#include "jackctrl.h"

extern global_info gi;

ConfigPortColors::ConfigPortColors(GUIconfig *p) : QGroupBox("Setup port colors:", p), parent(p), isHandlingSelChg(false) {
	layout = new QGridLayout(this, 8, 3, 10);
	note1 = new QLabel("Doubleclick color to change it.", this);
	note2 = new QLabel("Ports with same color will be grouped together when quick selecting outputs.", this);
	availPorts = new QListBox(this);
	mappedPorts = new QListBox(this);
	button = new QPushButton(" =!= ", this);
	labelAvail = new QLabel("Available ports:", this);
	labelMapped = new QLabel("Mapped ports:", this);

	QFont fontHeading;
	fontHeading.setWeight(QFont::Bold);
	setFont(fontHeading);

	QFont fontOthers;
	fontOthers.setWeight(QFont::Normal);
	note1->setFont(fontOthers);
	note2->setFont(fontOthers);
	availPorts->setFont(fontOthers);
	mappedPorts->setFont(fontOthers);
	button->setFont(fontOthers);
	labelAvail->setFont(fontOthers);
	labelMapped->setFont(fontOthers);

	availPorts->setFixedSize(160, 132);
	mappedPorts->setFixedSize(160, 132);

	availPorts->setSelectionMode(QListBox::Extended);
	mappedPorts->setSelectionMode(QListBox::Extended);
	button->setEnabled(false);

	layout->setRowSpacing (0, 7);
	layout->addMultiCellWidget(availPorts, 2, 4, 0, 0);
	layout->addWidget(labelAvail, 1, 0, Qt::AlignCenter);
	layout->addWidget(labelMapped, 1, 2, Qt::AlignCenter);
	layout->addWidget(button, 3, 1);
	layout->addMultiCellWidget(mappedPorts, 2, 4, 2, 2);
	layout->addMultiCellWidget(note1, 5, 5, 0, 2, Qt::AlignRight);
	layout->addMultiCellWidget(note2, 6, 6, 0, 2, Qt::AlignRight);
	layout->setRowSpacing (7, 5);

	setSizePolicy(QSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum));

	connect(button, SIGNAL(clicked()), this, SLOT(click()));
	connect(availPorts, SIGNAL(selectionChanged()), this, SLOT(sigAvailSelChanged()));
	connect(mappedPorts, SIGNAL(selectionChanged()), this, SLOT(sigMappedSelChanged()));
	connect(mappedPorts, SIGNAL(doubleClicked(QListBoxItem *)), this, SLOT(mappedDoubleClicked(QListBoxItem *)));
}

ConfigPortColors::~ConfigPortColors() {
	delete note1;
	delete note2;
	delete labelAvail;
	delete labelMapped;
	delete availPorts;
	delete mappedPorts;
	delete button;
	delete layout;
}

void ConfigPortColors::getData() {
	char **itPorts;
	PortColorList::iterator itList;
	QListBoxItem *boxitem;

	availPorts->clear();
	mappedPorts->clear();


	int colIndex = 0;
	mappedPorts->insertItem(new ListBoxPixmapExt(QString(" - Unknown ports - "), parent->unknownPort, false));

	// Update available ports
	port_info *jackports = gi.jackPortsInfo;
	if(jackports) {
		itPorts = (char **) jackports->names;
		while(*itPorts) {                                   // For each port name
			itList = parent->listPortColor.begin();
			while(itList != parent->listPortColor.end()) {  // Search if we know it
				if((*itList).name == *itPorts)
					break;
				itList++;
			}

			if(itList == parent->listPortColor.end())       // Port name not known - add to available
				availPorts->insertItem(new ListBoxTextExt(QString(*itPorts), true));
			else {											// Add to know ports
				mappedPorts->insertItem(new ListBoxPixmapExt((*itList).name, (*itList).color), true);

				if(colIndex < QColorDialog::customCount() && !hasColor((*itList).color, colIndex)) {      // Add color to dialog
					QColorDialog::setCustomColor(colIndex, (*itList).color.rgb());
					colIndex++;
				}
			}

			itPorts++;
		}
	}
	else
		QMessageBox::warning(this, "Jack communication", "An error happened while requesting information about jack's ports" , QMessageBox::Ok, QMessageBox::NoButton);


	// Update known ports
	itList = parent->listPortColor.begin();
	while(itList != parent->listPortColor.end()) {
		boxitem = mappedPorts->findItem((*itList).name, Qt::CaseSensitive | Qt::ExactMatch);
		if(boxitem == 0) {
			mappedPorts->insertItem(new ListBoxPixmapExt((*itList).name, (*itList).color, false));

			if(colIndex < QColorDialog::customCount() && !hasColor((*itList).color, colIndex)) {      // Add color to dialog
				QColorDialog::setCustomColor(colIndex, (*itList).color.rgb());
				colIndex++;
			}
		}

		itList++;
	}

	// Insert default colors into colordialog
	int i = 0;
	QColor col;
	while(colIndex < QColorDialog::customCount() && i < DEF_COLOR_NUM) {
		col = getColor(i);
		if(!hasColor(col, colIndex)) {
			QColorDialog::setCustomColor(colIndex, col.rgb());
			colIndex++;
		}
		i++;
	}
}

bool ConfigPortColors::hasColor(QColor col, int index) {
	for(int i = 0; i < index; i++)
		if(col == QColor(QColorDialog::customColor(i)))
			return true;

	return false;
}

void ConfigPortColors::saveData() {
	ListBoxPixmapExt *ptr;
	portVScolor entry;
	parent->listPortColor.clear();

	if(mappedPorts->count() == 0)  // Error should at least have the default entry
		return; 

	ptr = (ListBoxPixmapExt *) mappedPorts->item(0);
	parent->unknownPort = ptr->color;
	for(uint i = 1; i < mappedPorts->count(); i++) {
		ptr = (ListBoxPixmapExt *) mappedPorts->item(i);
		entry.color = ptr->color;
		entry.name = mappedPorts->item(i)->text();
		parent->listPortColor.push_back(entry);
	}

	qHeapSort(parent->listPortColor);
}

QColor ConfigPortColors::getNextColor() {
	uint i;
	if((i = mappedPorts->count())) {
		i--;       // The default color
		i /= 2;    // Group the two by two
		i %= DEF_COLOR_NUM;

		return getColor(i);
	}

	return QColor(DEF_COLOR_PORT_0);
}

void ConfigPortColors::updateButton() {
	if(canAdd && !canRem) {
		button->setText("Add =>");
		button->setEnabled(true);
		return;
	}

	if(!canAdd && canRem) {
		button->setText("Remove <=");
		button->setEnabled(true);
		return;
	}

	if(!canAdd && !canRem) {
		button->setText(" =!= ");
		button->setEnabled(false);
		return;
	}

	button->setText("ERROR");
	button->setEnabled(false);
}

void ConfigPortColors::click() {
	if(canAdd) {
		doAdd();
		return;
	}

	if(canRem)
		doRem();
}

QColor ConfigPortColors::getColor(int i) {
		switch(i) {
			case 1:
				return QColor(DEF_COLOR_PORT_1);
			case 2:
				return QColor(DEF_COLOR_PORT_2);
			case 3:
				return QColor(DEF_COLOR_PORT_3);
			case 4:
				return QColor(DEF_COLOR_PORT_4);
			case 5:
				return QColor(DEF_COLOR_PORT_5);
			case 6:
				return QColor(DEF_COLOR_PORT_6);
			case 7:
				return QColor(DEF_COLOR_PORT_7);
			case 8:
				return QColor(DEF_COLOR_PORT_8);
			case 9:
				return QColor(DEF_COLOR_PORT_9);
			default:   // == color 0
				return QColor(DEF_COLOR_PORT_0);
		}
}

void ConfigPortColors::doAdd() {
	QListBoxItem *ptr;
	QColor col;

	if(availPorts->count()) {
		uint index = 0;
		while(index < availPorts->count()) {
			ptr = availPorts->item(index);
			if(ptr->isSelected()) {
				mappedPorts->insertItem(new ListBoxPixmapExt(ptr->text(), getNextColor(), ((ListBoxTextExt *) ptr)->active));
				availPorts->removeItem(index);
			}
			else
				index++;
		}
	}

	canAdd = false;
}

void ConfigPortColors::doRem() {
	QListBoxItem *ptr;
	QColor col;

	if(mappedPorts->count()) {
		uint index = 0;
		while(index < mappedPorts->count()) {
			ptr = mappedPorts->item(index);
			if(ptr->isSelected()) {
				availPorts->insertItem(new ListBoxTextExt(ptr->text(), ((ListBoxPixmapExt *) ptr)->active));
				mappedPorts->removeItem(index);
			}
			else
				index++;
		}
	}

	canRem = false;
}

void ConfigPortColors::sigMappedSelChanged() {
	if(isHandlingSelChg)
		return;
	isHandlingSelChg = true;

	if(mappedPorts->count() == 0) {
		isHandlingSelChg = false;
		return;
	}

	QListBoxItem *ptr = mappedPorts->firstItem();
	uint index = 0;
	availPorts->clearSelection();
	canAdd = false;
	if(!(ptr->isSelected())) {  // Default (number 1) may not be selected
		index++;
		ptr = mappedPorts->item(index);
		while(index < mappedPorts->count()) {
			if(ptr->isSelected()) { // Another one selected
				canRem = true;
				isHandlingSelChg = false;
				updateButton();
				return;
			}
			index++;
			ptr = mappedPorts->item(index);
		}
	}

	canRem = false;
	updateButton();
	isHandlingSelChg = false;
}

void ConfigPortColors::sigAvailSelChanged() {
	if(isHandlingSelChg)
		return;
	isHandlingSelChg = true;

	if(availPorts->count() == 0) {
		isHandlingSelChg = false;
		return;
	}

	uint index = 0;
	QListBoxItem *ptr = availPorts->firstItem();
	mappedPorts->clearSelection();
	canRem = false;

	while(index < availPorts->count()) {
		if(ptr->isSelected()) {
			canAdd = true;
			isHandlingSelChg = false;
			updateButton();
			return;
		}
		index++;
		ptr = availPorts->item(index);
	}

	canAdd = false;
	isHandlingSelChg = false;
	updateButton();
}

void ConfigPortColors::mappedDoubleClicked(QListBoxItem * item) {
	QColor c = QColorDialog::getColor(((ListBoxPixmapExt *) item)->color, this,0);

	if(c.isValid())
		((ListBoxPixmapExt *) item)->updateColor(c);
}


// ************************************************************************************************

ConfigKeys::ConfigKeys(GUIconfig *p) : QGroupBox("Setup keys:", p), parent(p) {
	layout = new QGridLayout(this, 4, 1, 10, 7);

	note1 = new QLabel("Doubleclick action to assign new key.", this);
	listKeys = new QListBox(this);

	QFont fontHeading;
	fontHeading.setWeight(QFont::Bold);
	setFont(fontHeading);

	QFont fontOthers;
	fontOthers.setWeight(QFont::Normal);
	note1->setFont(fontOthers);
	listKeys->setFont(fontOthers);

	layout->setRowSpacing (0, 7);
	layout->addWidget(listKeys, 1, 0);
	layout->addWidget(note1, 2, 0);
	layout->setRowSpacing (3, 5);

	connect(listKeys, SIGNAL(doubleClicked(QListBoxItem *)), this, SLOT(clicked(QListBoxItem *)));
	connect(listKeys, SIGNAL(returnPressed(QListBoxItem *)), this, SLOT(clicked(QListBoxItem *)));
}

void ConfigKeys::getData() {
	k_ctrl = parent->key_ctrl;
	k_play = parent->key_play_next;
	k_skip = parent->key_skip;
	k_alloff = parent->key_alloff;
	k_kill1 = parent->key_kill1;
	k_kill2 = parent->key_kill2;
	k_kill3 = parent->key_kill3;
	k_kill4 = parent->key_kill4;
	k_kill5 = parent->key_kill5;
	k_kill6 = parent->key_kill6;
	k_kill7 = parent->key_kill7;
	k_kill8 = parent->key_kill8;
	k_kill9 = parent->key_kill9;
	k_kill10 = parent->key_kill10;
	k_kill11 = parent->key_kill11;
	k_kill12 = parent->key_kill12;

	updateListbox();
}

void ConfigKeys::saveData() {
	parent->key_ctrl = k_ctrl;
	parent->key_play_next = k_play;
	parent->key_skip = k_skip;
	parent->key_alloff = k_alloff;
	parent->key_kill1 = k_kill1;
	parent->key_kill2 = k_kill2;
	parent->key_kill3 = k_kill3;
	parent->key_kill4 = k_kill4;
	parent->key_kill5 = k_kill5;
	parent->key_kill6 = k_kill6;
	parent->key_kill7 = k_kill7;
	parent->key_kill8 = k_kill8;
	parent->key_kill9 = k_kill9;
	parent->key_kill10 = k_kill10;
	parent->key_kill11 = k_kill11;
	parent->key_kill12 = k_kill12;
}

void ConfigKeys::updateListbox() {
	listKeys->clear();

	listKeys->insertItem(new ListBoxTextNum("Modifier: " + key2string(k_ctrl), ctrl));
	listKeys->insertItem(new ListBoxTextNum("Play next: " + key2string(k_play), play));
	listKeys->insertItem(new ListBoxTextNum("Skip next: " + key2string(k_skip), skipnext));
	listKeys->insertItem(new ListBoxTextNum("All off: " + key2string(k_alloff), alloff));
	listKeys->insertItem(new ListBoxTextNum("Kill #1: " + key2string(k_kill1), kill1));
	listKeys->insertItem(new ListBoxTextNum("Kill #2: " + key2string(k_kill2), kill2));
	listKeys->insertItem(new ListBoxTextNum("Kill #3: " + key2string(k_kill3), kill3));
	listKeys->insertItem(new ListBoxTextNum("Kill #4: " + key2string(k_kill4), kill4));
	listKeys->insertItem(new ListBoxTextNum("Kill #5: " + key2string(k_kill5), kill5));
	listKeys->insertItem(new ListBoxTextNum("Kill #6: " + key2string(k_kill6), kill6));
	listKeys->insertItem(new ListBoxTextNum("Kill #7: " + key2string(k_kill7), kill7));
	listKeys->insertItem(new ListBoxTextNum("Kill #8: " + key2string(k_kill8), kill8));
	listKeys->insertItem(new ListBoxTextNum("Kill #9: " + key2string(k_kill9), kill9));
	listKeys->insertItem(new ListBoxTextNum("Kill #10: " + key2string(k_kill10), kill10));
	listKeys->insertItem(new ListBoxTextNum("Kill #11: " + key2string(k_kill11), kill11));
	listKeys->insertItem(new ListBoxTextNum("Kill #12: " + key2string(k_kill12), kill12));
}

void ConfigKeys::setKey(FuncKeys f, int k) {
	switch(f) {
		case ctrl:
			k_ctrl = k;
			break;
		case play:
			k_play = k;
			break;
		case skipnext:
			k_skip = k;
			break;
		case alloff:
			k_alloff = k;
			break;
		case kill1:
			k_kill1 = k;
			break;
		case kill2:
			k_kill2 = k;
			break;
		case kill3:
			k_kill3 = k;
			break;
		case kill4:
			k_kill4 = k;
			break;
		case kill5:
			k_kill5 = k;
			break;
		case kill6:
			k_kill6 = k;
			break;
		case kill7:
			k_kill7 = k;
			break;
		case kill8:
			k_kill8 = k;
			break;
		case kill9:
			k_kill9 = k;
			break;
		case kill10:
			k_kill10 = k;
			break;
		case kill11:
			k_kill11 = k;
			break;
		case kill12:
			k_kill12 = k;
			break;
	}

	updateListbox();
}

void ConfigKeys::show() {
	updateListbox();
	QWidget::show();
}

void ConfigKeys::clicked(QListBoxItem *i) {
	parent->grabKey->show(((ListBoxTextNum *) i)->fk);
}

// ************************************************************************************************

GrabKey::GrabKey(ConfigKeys *dp, QWidget *pp) : 
	QLabel("Press new key or wait\n1.5 sec to select none.", pp, 0, Qt::WStyle_NoBorder | Qt::WStyle_StaysOnTop | Qt::WType_Popup),
	dataParent(dp), parent(pp) {
	setMinimumSize(200, 150);
	timer = new QTimer(this);
	setAlignment(Qt::AlignCenter);
	connect(timer, SIGNAL(timeout()), this, SLOT(cancel()));
	setFrameStyle(QFrame::Sunken | QFrame::Panel);
}

void GrabKey::show(FuncKeys k) {
	timer->start(TIMER_KEY_DELAY, true);
	key = k;
	QLabel::show();
	move(parent->x() + (parent->sizeHint().width() - width()) / 2, 
		 parent->y() + (parent->sizeHint().height() - height()) / 2);
}

void GrabKey::keyPressEvent(QKeyEvent *e) {
	timer->stop();
	dataParent->setKey(key, e->key());
	hide();
}

void GrabKey::cancel() {
	dataParent->setKey(key, 0);
	hide();
}

// ************************************************************************************************

MiscOpts::MiscOpts(GUIconfig *p) : QGroupBox("Misceleneous options:", p), parent(p){
	setColumnLayout(2, Qt::Horizontal);
	checkSaveOnClose = new QCheckBox("Save on close", this);
	checkShowRemain = new QCheckBox("Show time remaining", this);
}

void MiscOpts::getData() {
	checkSaveOnClose->setChecked(parent->saveOnClose);
	checkShowRemain->setChecked(parent->showRemaining);
}

void MiscOpts::saveData() {
	parent->saveOnClose = checkSaveOnClose->isChecked();
	parent->showRemaining = checkShowRemain->isChecked();
}
