#include <qapplication.h>

#include "x11routines.h"
#include "guiconfig.h"
#include "config.h"
#include "jackctrl.h"

extern global_info gi;

GUIconfig::GUIconfig() :
	QWidget(0, "Preferences", WType_Dialog | WStyle_DialogBorder | WShowModal),
	parent(NULL) {

	setCaption("TTtrigger - Preferences");

	setDefaults();
	fileName = QString(DEF_FILE_NAME);  // Should not be in setDefaults(...)

	setupWidgets();
}

GUIconfig::~GUIconfig() {
	delete buttonOK;
	delete buttonCancel;
	delete widgetPortColors;
	delete widgetMisc;
	delete grabKey;
	delete configKeys;
	delete layout;
}

void GUIconfig::show() {
	setupData();

	QWidget::show();

	X11Rt::setClassName(this, QString("Preferences"), QString("TTtrigger"));
}

void GUIconfig::setupWidgets() {
	layout = new QGridLayout(this, 3, 3, 15);

	configKeys = new ConfigKeys(this);
	grabKey = new GrabKey(configKeys, this);
	buttonOK = new QPushButton("OK", this);
	buttonCancel = new QPushButton("Cancel", this);
	widgetPortColors = new ConfigPortColors(this);
	widgetMisc = new MiscOpts(this);

	layout->addMultiCellWidget(widgetPortColors, 0, 0, 0, 1);
	layout->addWidget(configKeys, 0, 2);
	layout->addMultiCellWidget(widgetMisc, 1, 1, 0, 1);
	layout->addWidget(buttonOK, 2, 0, Qt::AlignLeft);
	layout->addWidget(buttonCancel, 2, 2, Qt::AlignRight);

	connect(buttonOK, SIGNAL(clicked()), this, SLOT(clickOK()));
	connect(buttonCancel, SIGNAL(clicked()), this, SLOT(hide()));

	setSizePolicy(QSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum));
	layout->setResizeMode(QLayout::Fixed);

}

void GUIconfig::setupData() {
	widgetPortColors->getData();
	configKeys->getData();
	widgetMisc->getData();
}

void GUIconfig::clickOK() {
	widgetPortColors->saveData();
	configKeys->saveData();
	widgetMisc->saveData();
	parent->menu_file->setItemChecked(parent->saveOnCloseId, saveOnClose);

	parent->playlist->updateGraphics();

	hide();
}

void GUIconfig::setDefaults() {
	saveOnClose = DEF_SAVE_ON_CLOSE;
	unknownPort.setRgb(DEF_COLOR_UNKNOWN_PORT);
	listPortColor.clear();

	key_ctrl = DEF_CTRL_MOD;
	key_play_next = DEF_KEY_PLAY_NEXT;
	key_skip = DEF_KEY_SKIP;
	key_alloff = DEF_KEY_ALLOFF;
	key_kill1 = DEF_KEY_KILL_1;
	key_kill2 = DEF_KEY_KILL_2;
	key_kill3 = DEF_KEY_KILL_3;
	key_kill4 = DEF_KEY_KILL_4;
	key_kill5 = DEF_KEY_KILL_5;
	key_kill6 = DEF_KEY_KILL_6;
	key_kill7 = DEF_KEY_KILL_7;
	key_kill8 = DEF_KEY_KILL_8;
	key_kill9 = DEF_KEY_KILL_9;
	key_kill10 = DEF_KEY_KILL_10;
	key_kill11 = DEF_KEY_KILL_11;
	key_kill12 = DEF_KEY_KILL_12;

	showRemaining = false;
}

void GUIconfig::keyPressEvent (QKeyEvent *e) {
	if (e->ascii() == 27)
		buttonCancel->animateClick();
	else if (e->ascii() == 13)
		buttonOK->animateClick();
	else
		e->ignore();
}

bool GUIconfig::load() {
	QString iLine, key, valueString;
	QStringList valueList;
	uint lineNr = 0;
	int i;
	bool parsed;
	entryType type = None;
	ChannelPortList chnVSport;
	QString audioFilename, audioName;
	int audioHotkey, audioVolume, repeatCount;

	if(!file.exists()) {
		qWarning("File %s does not exist.\nWill be created when saveing.\nUsing defaults.", file.name().ascii());
		return true;
	}

	if (!file.open(IO_ReadOnly)) {
		qWarning("Could not open file: %s.", file.name().ascii());
		return false;
	}

	if(gi.debugLevel > 1)
		qWarning("Reading file %s.", file.name().ascii());

	while (file.readLine(iLine, 1024) != -1) {
		lineNr++;
		parsed = false;
		iLine = iLine.stripWhiteSpace();
		if(!iLine.isEmpty()) {
			// Comments
			if(iLine.at(0) == '#')
				parsed = true;		
			else if(iLine.at(0) == '[' && iLine.at(iLine.length() - 1) == ']') {
				if(type == Audio)
					playlist->openAddEntry(audioFilename, audioName, chnVSport, audioHotkey, audioVolume, repeatCount);
				else if(type == Sep)
					playlist->insertSep(audioName);

				audioFilename = iLine.mid(1, iLine.length() - 2);
				if(audioFilename == "-Separator-")
					type = Sep;
				else
					type = Audio;

				audioName = QString();
				chnVSport.clear();
				audioHotkey = 0;
				repeatCount = 1;

				parsed = true;
			}

			else if(iLine.at(0) == '>') {
				if(type == None)
					parsed = parseConfig(&iLine, &listPortColor);
				else if(type == Audio)
					parsed = parseConfig(&iLine, &chnVSport);
			}

			else if ((i = iLine.find('=')) != -1) {
				key = iLine.left(i).stripWhiteSpace().lower();
				valueString = iLine.right(iLine.length() - i - 1).stripWhiteSpace();
				valueList = QStringList::split(QChar(' '), valueString);
				if(!(key.isEmpty() || valueString.isEmpty())) {
					if(type != None) {
						if(key == "name") {
							parsed = true;
							audioName = valueString;
						}
						else if(key == "hotkey") {
							parsed = true;
							parseConfig(&valueString, &audioHotkey);
						}
						else if(key == "volume") {
							parsed = true;
							parseConfig(&valueString, &audioVolume);
						}
						else if(key == "repeats") {
							parsed = true;
							parseConfig(&valueString, &repeatCount);
						}
					}
					else {
						if(key == "saveonexit")
							parsed = parseConfig(&valueString, &saveOnClose);
						if(key == "showremaining")
							parsed = parseConfig(&valueString, &showRemaining);
						else if(key == "unknownportcolor")
							parsed = parseConfig(&valueString, &valueList, &unknownPort);
						else if(key == "key_ctrl")
							parsed = parseConfig(&valueString, &key_ctrl);
						else if(key == "key_play_next")
							parsed = parseConfig(&valueString, &key_play_next);
						else if(key == "key_skip")
							parsed = parseConfig(&valueString, &key_skip);
						else if(key == "key_alloff")
							parsed = parseConfig(&valueString, &key_alloff);
						else if(key == "key_kill1")
							parsed = parseConfig(&valueString, &key_kill1);
						else if(key == "key_kill2")
							parsed = parseConfig(&valueString, &key_kill2);
						else if(key == "key_kill3")
							parsed = parseConfig(&valueString, &key_kill3);
						else if(key == "key_kill4")
							parsed = parseConfig(&valueString, &key_kill4);
						else if(key == "key_kill5")
							parsed = parseConfig(&valueString, &key_kill5);
						else if(key == "key_kill6")
							parsed = parseConfig(&valueString, &key_kill6);
						else if(key == "key_kill7")
							parsed = parseConfig(&valueString, &key_kill7);
						else if(key == "key_kill8")
							parsed = parseConfig(&valueString, &key_kill8);
						else if(key == "key_kill9")
							parsed = parseConfig(&valueString, &key_kill9);
						else if(key == "key_kill10")
							parsed = parseConfig(&valueString, &key_kill10);
						else if(key == "key_kill11")
							parsed = parseConfig(&valueString, &key_kill11);
						else if(key == "key_kill12")
							parsed = parseConfig(&valueString, &key_kill12);
					}
				}
			}
			
			if (!parsed)
				qWarning("Could not parse \"%s\" at line %d.", iLine.ascii(), lineNr);
		}
	}

	file.close();

	if(type == Audio)
		playlist->openAddEntry(audioFilename, audioName, chnVSport, audioHotkey, audioVolume, repeatCount);
	else if(type == Sep)
		playlist->insertSep(audioName);

	qHeapSort(listPortColor);

	return true;
}

bool GUIconfig::save(bool startTimers){
	if (!file.open(IO_WriteOnly)) {
		if(gi.debugLevel > 0)
			qWarning("ERROR: Could open file %s for output!", file.name().ascii());

		return false;
	}

	// Stop timers
	for(PlaylistBase *wit = playlist->itemFirst; wit; wit = wit->next)
		if(wit->isEntry)
			((PlaylistEntry *) wit)->timer->stop();
	grabKey->timer->stop();
	((MainWindow *) qApp->mainWidget())->activeWindow->aPlaylist->timersStop();
	

	QTextStream stream( &file );

	stream << "SaveOnExit = " << (saveOnClose ? 1 : 0) << "\n";
	stream << "ShowRemaining = " <<(showRemaining ? 1 : 0) << "\n";
	stream << "UnknownPortColor = " << unknownPort.name() << "\n";

	stream << "\n";

	PortColorList::iterator it = listPortColor.begin();
	while(it != listPortColor.end()) {
	stream << ">" << (*it).name << "< " << (*it).color.name() << "\n";
		it++;
	}

	stream << "\n";

	stream << "key_ctrl = " << key_ctrl << "\n" <<
		"key_play_next = " << key_play_next << "\n" <<
		"key_skip = " << key_skip << "\n" <<
		"key_alloff = " << key_alloff << "\n" <<
		"key_kill1 = " << key_kill1 << "\n" <<
		"key_kill2 = " << key_kill2 << "\n" <<
		"key_kill3 = " << key_kill3 << "\n" <<
		"key_kill4 = " << key_kill4 << "\n" <<
		"key_kill5 = " << key_kill5 << "\n" <<
		"key_kill6 = " << key_kill6 << "\n" <<
		"key_kill7 = " << key_kill7 << "\n" <<
		"key_kill8 = " << key_kill8 << "\n" <<
		"key_kill9 = " << key_kill9 << "\n" <<
		"key_kill10 = " << key_kill10 << "\n" <<
		"key_kill11 = " << key_kill11 << "\n" <<
		"key_kill12 = " << key_kill12 << "\n";

	stream << "\n";

	for(PlaylistBase *wit = playlist->itemFirst; wit; wit = wit->next)
		stream << wit->saveText() << "\n";

	file.close();

	if(startTimers)
		((MainWindow *) qApp->mainWidget())->activeWindow->aPlaylist->timersStart();

	return true;
}

bool GUIconfig::parseConfig(QString * inp, PortColorList * out) {
	portVScolor entry;

	int i = inp->findRev('<');
	if(i < 2)
		return false;

	entry.name = inp->mid(1, i - 1);;
	if(entry.name.isEmpty())
		return false;

	QString colstr = inp->right(inp->length() - i - 1).stripWhiteSpace();
	QStringList collist = QStringList::split(QChar(' '), colstr);
	
	if(parseConfig(&colstr, &collist, &(entry.color))) {
		out->push_back(entry);
		return true;
	}

	return false;
}

bool GUIconfig::parseConfig(QString * inp, ChannelPortList * out) {
	channelVSport entry;

	int i = inp->find(' ');
	if(i < 2)
		return false;

	entry.channel = inp->mid(1, i - 1).toUInt();
	entry.name = inp->right(inp->length() - i - 1);
	if(entry.name.isEmpty())
		return false;

	out->push_back(entry);

	return true;
}

bool GUIconfig::parseConfig(QString * inp, bool * out) {
	*out = (*inp == "true" || *inp == "1");
	
	return true;

}

bool GUIconfig::parseConfig(QString * inp, QStringList * list, QColor * out) {
	int i;
	bool parsed, ok1, ok2, ok3;

	if ((i = list->count()) == 1) {
		if (parsed = (inp->at(0) == '#' && inp->length() == 7))
			out->setRgb((inp->mid(1,2)).toUInt(&ok1, 16),
						(inp->mid(3,2)).toUInt(&ok2, 16),
						(inp->mid(5,2)).toUInt(&ok3, 16));
	}
	else if(parsed = (i == 3))
		out->setRgb((*list->at(0)).toUInt(&ok1),
					(*list->at(1)).toUInt(&ok2),
					(*list->at(2)).toUInt(&ok3));
	
	return parsed && ok1 && ok2 && ok3;
}

bool GUIconfig::parseConfig(QString *inp, QFont *out) {
	return out->fromString(*inp);
}

bool GUIconfig::parseConfig(QString *inp, int *out) {
	*out = inp->toInt();
	
	return true;
}

void GUIconfig::setParent(MainWindow *p, Playlist *pl) {
	parent = p;
	playlist = pl;
	reparent(p, getWFlags(), pos(), isShown());
}
