#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <string.h>

#include "jackctrl.h"

/********* int init_jackctrl(global_info* gi) **********
*  This function should
*  - create a jack-connection
*  - create the process thread (really done by jack)
*  - create disk thread process
*  - initialize read/write lock  (linked audio list; obtain write lock before deleting entry)
*  - initialize condition/mutex  (signal jproc => disk_read - requesting data)
*  - return 0 on success
*******************************************************/
int init_jackctrl(global_info *gi) {
	if(gi == NULL) {
		fprintf (stderr, "Missing argument for init_jackctrl!\n");
		return -1;
	}

	gi->client = 0;
	gi->internalInterlacedBuffer = 0;
	gi->internalBufferArray = 0;
	gi->internalJackBuffer = 0;
	gi->jack_samplerate = 0;
	gi->audioList = 0;
	gi->audioIdCounter = 0;
	gi->jackPortsInfo = 0;
	gi->jackPorts = 0;
	gi->audioSampleSize = sizeof(jack_default_audio_sample_t);

	gi->thread_status = 0;

	if(gi->clientName == NULL) {
		fprintf (stderr, "No client name specified!\n");
		return -1;
	}

	if((gi->client = jack_client_new(gi->clientName)) == NULL) {
		fprintf (stderr, "Could not connect to jack.\nThis can be caused by several things:\n");
		fprintf (stderr, "Jack server not running?\n");
		fprintf (stderr, "Jack server running as another user?\n");
		fprintf (stderr, "Client with name %s already registered?\n", gi->clientName);
		return -1;
	}

	// This func does not return any status like the other funcs. Dont why...?
	jack_on_shutdown(gi->client, jack_shutdown, gi);

	if(jack_set_process_callback(gi->client, jprocess, gi)) {
		fprintf (stderr, "Could not set process callback function.\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}

	if(gi->debugLevel > 0)
		if(jack_is_realtime(gi->client) == 0) {
			fprintf(stderr, "WARNING: Jack server not running realtime mode.\n");
			fprintf(stderr, "         You would probably be getting a better performance by\n");
			fprintf(stderr, "         starting jackd with realtime mode (-R option).\n");
		}

	if(jack_set_sample_rate_callback(gi->client, set_samplerate, gi)) {
		fprintf (stderr, "Could not set sample rate callback function\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}
	
	gi->jackBufSize = jack_get_buffer_size(gi->client);

	if(gi->debugLevel > 1)
		fprintf(stderr, "Requesting internal interlaced buffer size of %d\n", gi->internalBufferSize * gi->maxAudioChannels);
	if((gi->internalInterlacedBuffer = (char*) malloc(gi->internalBufferSize * gi->maxAudioChannels)) == NULL) {
		fprintf (stderr, "Could not initialize internal buffers (interlaced).\n");
		fprintf (stderr, "Make sure you have enough system ressources available.\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}

	if(gi->debugLevel > 1)
		fprintf(stderr, "Requesting internal buffer size of %d\n", gi->internalBufferSize * gi->maxAudioChannels);
	if((gi->internalBufferArray = (char*) malloc(gi->internalBufferSize * gi->maxAudioChannels)) == NULL) {
		fprintf (stderr, "Could not initialize internal buffers (array).\n");
		fprintf (stderr, "Make sure you have enough system ressources available.\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}

	if(gi->debugLevel > 1)
		fprintf(stderr, "Requesting internal jack buffer size of %d * %d = %d\n", gi->jackBufSize, gi->audioSampleSize, gi->jackBufSize * gi->audioSampleSize);
	if((gi->internalJackBuffer = (jack_default_audio_sample_t *) malloc(gi->jackBufSize * gi->audioSampleSize)) == NULL) {
		fprintf (stderr, "Could not initialize internal buffers (jack).\n");
		fprintf (stderr, "Make sure you have enough system ressources available.\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}

	if(jack_set_buffer_size_callback(gi->client, set_jackbufsize, gi)) {
		fprintf (stderr, "Could not set buffer_size_change callback function\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}

	if(pthread_cond_init(&(gi->signal_condition), NULL)) {
		fprintf (stderr, "Could not initialize internal signal condition (pthreads).\n");
		fprintf (stderr, "Make sure you have enough system ressources available.\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}
	gi->thread_status |= THREAD_STATUS_SIG_COND;

	pthread_mutex_init(&(gi->condition_mutex), NULL);   // Always returns 0
	gi->thread_status |= THREAD_STATUS_COND_MUT;

	pthread_rwlock_init(&(gi->rwlock), NULL);
	gi->thread_status |= THREAD_STATUS_RWLOCK;

	if(pthread_create(&(gi->pthread), NULL, disk_read, gi)) {
		fprintf(stderr, "Could not launch new thread.\n");
		fprintf(stderr, "Make sure you have enough system ressources available.\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}
	gi->thread_status |= THREAD_STATUS_PTHREAD;

	gi->audioSampleSize = sizeof(jack_default_audio_sample_t);

	if (jack_activate (gi->client)) {
		fprintf (stderr, "Could not activate client.\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}

	setupPorts(gi);
	
	return 0;
}

/******** void stop_jackctrl(global_info *gi) **********
*  This function cleans up after init_jackctrl(...)
*******************************************************/
void stop_jackctrl(global_info *gi) {
	int i;
	
	if(gi->debugLevel > 1)
		fprintf(stderr, "INFO: Cleaning up... ");

	if(gi->thread_status & THREAD_STATUS_PTHREAD) {
		pthread_cancel(gi->pthread);
		pthread_join(gi->pthread, NULL);
	}

	audio_entry *delEntry, *entry = gi->audioList;
	while(entry) {
		delEntry = entry;
		entry = entry->next;
		destroy_audio_entry(delEntry);
	}
	gi->audioList = 0;

	if(gi->client) {
		if(gi->jackPorts) {
			i = 0;
			while(*(gi->jackPorts + i)) {
				jack_port_unregister(gi->client, *(gi->jackPorts + i));
				i++;
			}
			
			free(gi->jackPorts);
			gi->jackPorts = 0;
		}

//		jack_deactivate(gi->client);
		jack_client_close(gi->client);
		gi->client = 0;
	}

	if(gi->thread_status & THREAD_STATUS_RWLOCK)
		pthread_rwlock_destroy(&(gi->rwlock));
	if(gi->thread_status & THREAD_STATUS_COND_MUT)
		pthread_mutex_destroy(&(gi->condition_mutex));
	if(gi->thread_status & THREAD_STATUS_SIG_COND)
		pthread_cond_destroy(&(gi->signal_condition));

	gi->thread_status = 0;

	if(gi->internalInterlacedBuffer) {
		free(gi->internalInterlacedBuffer);
		gi->internalInterlacedBuffer = 0;
	}

	if(gi->internalBufferArray) {
		free(gi->internalBufferArray);
		gi->internalBufferArray = 0;
	}
	
	if(gi->internalJackBuffer) {
		free(gi->internalJackBuffer);
		gi->internalJackBuffer = 0;
	}

	i = 0;
	if(gi->jackPortsInfo) {
		if((gi->jackPortsInfo)->names) {
			while(*(gi->jackPortsInfo->names + i)) {
				free(*(gi->jackPortsInfo->names + i));
				i++;
			}
			
			free((gi->jackPortsInfo)->names);
		}
		
		free(gi->jackPortsInfo);
		gi->jackPortsInfo = 0;
	}
	
	if(gi->debugLevel > 1)
		fprintf(stderr, "Done.\n");
}



/********* void setupPorts(global_info *gi) ***********
*  Finds available Jack ports.
*  For each, create our own port and connect them.
*******************************************************/
void setupPorts(global_info *gi) {
	const char **ports;
	int i = 0, j, portNameLength;
	jack_port_t **portPtr;
	char *buf;
	
	buf = (char *) malloc(portNameLength = jack_port_name_size());
	if(buf == NULL) {
		fprintf(stderr, "ERROR: Could not allocate memory!\n");
		return;
	}
	
	if(portNameLength > (i = strlen(gi->clientName) + 2))
		portNameLength -= i;
	else {
		fprintf(stderr, "ERROR: Client name too long!\n");
		free(buf);
		return;
	}
		
	if((ports = jack_get_ports(gi->client, NULL, NULL, JackPortIsInput)) == NULL) {
		if(gi->debugLevel > 0)
			fprintf(stderr, "WARNING: Could not get jack ports!\n");
		free(buf);
		return;
	}

	gi->jackPortsInfo = (port_info *) malloc(sizeof(port_info));
	if(gi->jackPortsInfo == NULL) {
		if(gi->debugLevel > 0)
			fprintf(stderr, "WARNING: Could not allocate memory for jackPortsInfo!\n");
		free(buf);
		return;
	}	

	// --------------------------------
	// Register ports
	// --------------
	if((gi->jackPortsInfo->names = malloc(JACK_MAX_NUM_PORTS * sizeof(char*)))) {
		for(i = 0; i < JACK_MAX_NUM_PORTS; i++)
			*(gi->jackPortsInfo->names + i) = 0;
				
		gi->jackPorts = malloc(sizeof(jack_port_t *) * JACK_MAX_NUM_PORTS);
		if(gi->jackPorts) {
			portPtr = gi->jackPorts;
			for(i = 0; i < JACK_MAX_NUM_PORTS; i++) {
				*portPtr = NULL;
				portPtr++;
			}


			// --------------------------------
			// Find and connect and register ports
			// --------------
			portPtr = gi->jackPorts;			
			i = 0;
			while(*(ports+i) != 0 && i < JACK_MAX_NUM_PORTS) {
				if(gi->debugLevel > 1)
					fprintf(stderr, "INFO: Port %d: %s\n", i, *(ports+i));

				if((*(gi->jackPortsInfo->names + i) = malloc(strlen(*(ports+i)) + 1)))
					strcpy(*(gi->jackPortsInfo->names + i), *(ports+i));
				else {
					fprintf(stderr, "ERROR: Could not allocate memory for port name!!\n");
					continue;
				}

				j = snprintf(buf, portNameLength, "Chn %d", i);
				if(j > portNameLength) {
					if(gi->debugLevel > 0)
						fprintf(stderr, "WARNING: Port name truncated to\n         %s.\n", buf);
				}

				*portPtr = jack_port_register(gi->client, buf, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput|JackPortIsTerminal, 0);
				if(*portPtr) {
					if(gi->debugLevel > 1)
						fprintf(stderr, "INFO: Registered port %s.\n", buf);
						
					if(jack_connect(gi->client, jack_port_name(*portPtr), *(ports+i)))
						if(gi->debugLevel > 0)
							fprintf(stderr, "ERROR: Could not connect port %s to %s!\n", jack_port_name(*portPtr), *(ports+i));
				}
				else {
					if(gi->debugLevel > 0)
						fprintf(stderr, "ERROR: Could not register port.\n");
				}

				portPtr++;
				i++;
			}

			gi->jackPortsInfo->count = i;
		}
		else {
			fprintf(stderr, "ERROR: Could not allocate memory for port names!!\n");
			free(gi->jackPortsInfo);
			gi->jackPortsInfo = 0;
		}
	}
	else
		fprintf(stderr, "ERROR: Could not allocate memory for port names!!\n");

	free(ports);
	free(buf);
}


/*********** void jack_shutdown(void *arg) ************
*  Called when jack shuts down.
*  What to do..?? What to dooo????
*******************************************************/
void jack_shutdown(void *arg) {
	global_info *gi = (global_info *) arg;

	if(gi->debugLevel > 0)
		fprintf(stderr, "WARNING: Jack shutting down or we have been to slow and\n"
						"         therefore disconnected.\n");
	gi->client = 0;
	stop_jackctrl(gi);		
}


/*** int jprocess(jack_nframes_t nframes, void *arg) ***
*  The jack callback function.
*  Shovel some data...
*******************************************************/
int jprocess(jack_nframes_t nframes, void *arg) {
	jack_port_t **jack_port;							// Pointer for iterating
	char *portName;
	port_connection *portConnectionPtr;
	uint maxFrames;
	jack_default_audio_sample_t *jack_portBuffer, *internaleBufPtr;
	audio_entry *entry;
	uint chnCnt, chnEOB;
	jack_ringbuffer_data_t readVector[2];
	uint i, j;
	jack_default_audio_sample_t *bufPtr, *vuPtr;
	float absVal;
	uint minAdvance;

	global_info *gi = (global_info *) arg;
	// Get read-lock (so no items in the audio-list will be deleted!)
	pthread_rwlock_rdlock(&(gi->rwlock));
	
	if(gi->jackBufSize < nframes) {
		fprintf(stderr, "gi->jackBufSize was smaller than nframes provided by Jack!\n");
		maxFrames = gi->jackBufSize;
	}
	else
		maxFrames = nframes;


	// Iterate over our registered ports
	jack_port = gi->jackPorts;
	j = 0;
	while(*jack_port) {
		// Clear port buffer
		bzero(gi->internalJackBuffer, gi->jackBufSize * gi->audioSampleSize);
		
		// Find port name and find all audio entrys connected to that port
		portName = *(gi->jackPortsInfo->names + j);
		entry = gi->audioList;
		while(entry) {
			if((portConnectionPtr = entry->portConnections)) {
				while(1) {
					if(entry->fillZero < 2 && entry->eob == 0) 
						if(entry->channels > portConnectionPtr->srcChannel) 
							if(strcmp(portConnectionPtr->dstName, portName) == 0)
								add2buffer(gi, entry, *(entry->ringbuffers + portConnectionPtr->srcChannel), maxFrames, entry->fillZero);

					if(portConnectionPtr->last)
						break;
					else
						portConnectionPtr++;
				}
			}

			entry = entry->next;
		}
		
		if((jack_portBuffer = (jack_default_audio_sample_t *) jack_port_get_buffer(*jack_port, nframes))) {
			internaleBufPtr = gi->internalJackBuffer;
			for(i  = 0; i < maxFrames; i++) {
				if(*internaleBufPtr > 1.0)
					*jack_portBuffer = 1.0;
				else if(*internaleBufPtr < -1.0)
					*jack_portBuffer = -1.0;
				else
					*jack_portBuffer = *internaleBufPtr;

				jack_portBuffer++;
				internaleBufPtr++;
			}
		}
		else {
			fprintf(stderr, "ERROR: Failed getting a portbuffer for audio entry %s, port %s.\n",
				entry->name, jack_port_name(*jack_port));
			continue;
		}
		
		j++;
		jack_port++;
	}


	entry = gi->audioList;
	while(entry) {
		if(entry->eob == 0) {
			if(entry->fillZero) {
				if(entry->fillZero == 1)
					entry->fillZero = 2;
				else {
					entry->eob = 1;
						if(gi->debugLevel > 1)
							fprintf(stderr, "INFO: Marking %s for deletion.\n", entry->name);
				}
			}
			else {
				chnEOB = 0;
				minAdvance = -1;
				for(chnCnt = 0; chnCnt < entry->channels; chnCnt++) {
					jack_ringbuffer_get_read_vector(*(entry->ringbuffers + chnCnt), readVector);

					if((readVector[0].len + readVector[1].len) == 0)
						chnEOB++;

					vuPtr = entry->vu + chnCnt;
					bufPtr = (jack_default_audio_sample_t *) readVector[0].buf;
					for(j = 0; j < readVector[0].len; j += gi->audioSampleSize) {
//						*vuPtr = 0;  Do not clear - GUI will do that when reading it!
						absVal = fabsf(*bufPtr);
						if(absVal > *vuPtr)
							*vuPtr = absVal;
						bufPtr++;
					}
					bufPtr = (jack_default_audio_sample_t *) readVector[1].buf;
					for(j = 0; j < readVector[1].len; j += gi->audioSampleSize) {
						absVal = fabsf(*bufPtr);
						if(absVal > *vuPtr)
							*vuPtr = absVal;
						bufPtr++;
					}

					j = maxFrames * gi->audioSampleSize;
					if(j > readVector[0].len + readVector[1].len)
						j = readVector[0].len + readVector[1].len;
					
					if(minAdvance > j)
						minAdvance = j;

					 jack_ringbuffer_read_advance(*(entry->ringbuffers + chnCnt), j);
				}


				// End-of-buffers - all audio played
				if(chnEOB == entry->channels && entry->eof != 0) {
					entry->eob = 1;
					if(gi->debugLevel > 1)
						fprintf(stderr, "INFO: EOB of %s.\n", entry->name);
				}



				// Update position and shift audio entry
				entry->currentFrame += minAdvance / gi->audioSampleSize;
			}
		}

		entry = entry->next;
	}
			

	// Think we are done, so release delloc and 
	// signal disk_thread that we are thirsty for more...
	pthread_rwlock_unlock(&(gi->rwlock));
	if(pthread_mutex_trylock(&(gi->condition_mutex)) == 0) {
		pthread_cond_signal(&(gi->signal_condition));
		pthread_mutex_unlock(&(gi->condition_mutex));
	}

	return 0;
}

/************** void add2buffer(...) ****************
*  Merge data from ringbuffer into internal
*  buffer. The internal buffer will later be copied
*  to jack.
*****************************************************/
void add2buffer(global_info *gi, audio_entry *entry, jack_ringbuffer_t *ringbuf, int frames, int fade) {
	int doWriteFrames, framesWritten;
	jack_ringbuffer_data_t readVector[2];
	jack_default_audio_sample_t damp = 1.0, delta;
	jack_default_audio_sample_t *srcPtr, *dstPtr;
	int i;

	doWriteFrames = jack_ringbuffer_read_space(ringbuf) / gi->audioSampleSize;

	if(doWriteFrames) {
		if(doWriteFrames > frames)
			doWriteFrames = frames;
		else
			if(entry->eof == 0)
				fprintf(stderr, "WARNING: Ringbuffer underrun!\n");

		if(doWriteFrames) {
			dstPtr = gi->internalJackBuffer;
			jack_ringbuffer_get_read_vector(ringbuf, readVector);
			 
			if(fade) {
				delta = damp / (jack_default_audio_sample_t) (doWriteFrames + 1);

				framesWritten = 0;
				srcPtr = (jack_default_audio_sample_t *) readVector[0].buf;
				i = readVector[0].len / gi->audioSampleSize;
				while(framesWritten < i && framesWritten < doWriteFrames) {
			 		*dstPtr += *srcPtr * damp;
					dstPtr++;
					srcPtr++;
					framesWritten++;
					damp -= delta;
				}
				doWriteFrames -= framesWritten;
				
				srcPtr = (jack_default_audio_sample_t *) readVector[1].buf;

				framesWritten = 0;
				i = readVector[1].len / gi->audioSampleSize;
				while(framesWritten < i && framesWritten < doWriteFrames) {
			 		*dstPtr += *srcPtr * damp;
					dstPtr++;
					srcPtr++;
					framesWritten++;
					damp -= delta;
				}
			}
			else {
				framesWritten = 0;
				srcPtr = (jack_default_audio_sample_t *) readVector[0].buf;
				i = readVector[0].len / gi->audioSampleSize;
				while(framesWritten < i && framesWritten < doWriteFrames) {
			 		*dstPtr += *srcPtr;
					dstPtr++;
					srcPtr++;
					framesWritten++;
				}
				doWriteFrames -= framesWritten;
				
				srcPtr = (jack_default_audio_sample_t *) readVector[1].buf;

				framesWritten = 0;
				i = readVector[1].len / gi->audioSampleSize;
				while(framesWritten < i && framesWritten < doWriteFrames) {
			 		*dstPtr += *srcPtr;
					dstPtr++;
					srcPtr++;
					framesWritten++;
				}
			}
		}
	}
	else if(entry->eof == 0)
		fprintf(stderr, "WARNING: Ringbuffer empty and file was not EOF!\n");
}


/******** int set_samplerate(jack_nframes_t nframes, void *arg) **********
*  This function is called whenever jack changes its samplerate.
*  It is also called once when connecting to jack.
**************************************************************************/
int set_samplerate(jack_nframes_t nframes, void *arg) {
	global_info *gi = (global_info *) arg;

	if(gi->jack_samplerate == 0) {
		if(gi->debugLevel > 1)
			fprintf(stderr, "INFO: Jack samplerate: %d\n", nframes);
	}
	else {
		if(gi->debugLevel > 0)
			fprintf(stderr, "WARNING: Jack is changing samplerate.\n");
	}

	gi->jack_samplerate = nframes;

	return 0;
}

/******** int set_jackbufsize(jack_nframes_t nframes, void *arg) **********
*  This function is called whenever jack changes its port buffer size.
**************************************************************************/
int set_jackbufsize(jack_nframes_t nframes, void *arg) {
	global_info *gi = (global_info *) arg;

	if(gi->debugLevel > 0)
		fprintf(stderr, "WARNING: Jack is changing its buffer size to %d.\n", nframes);

	gi->jackBufSize = nframes;

	if(gi->internalJackBuffer) {
		free(gi->internalJackBuffer);
		gi->internalJackBuffer = 0;
	}

	if((gi->internalJackBuffer = (jack_default_audio_sample_t *) malloc(gi->jackBufSize * gi->audioSampleSize)) == NULL) {
		fprintf (stderr, "Could not initialize internal buffers (jack).\n");
		fprintf (stderr, "Make sure you have enough system ressources available.\n");
		fprintf (stderr, "Disconnecting...\n");
		jack_client_close(gi->client);
		return -1;
	}

	return 0;
}


/********** void *disk_read(void *arg) ************
*  PThread that takes care of shoveling data from
*  disk to ringbuffers.
***************************************************/
void *disk_read(void *arg) {
	audio_entry *entry, *lastEntry = NULL, *delEntry;
	int i;

	global_info *gi = (global_info *) arg;

	// Lock signal/condition mutex
	pthread_mutex_lock(&(gi->condition_mutex));

	while(1) {
		i = pthread_cond_wait(&(gi->signal_condition), &(gi->condition_mutex));

		// Loop though list and delete eob entrys
		entry = gi->audioList;
		lastEntry = NULL;

		while(entry) {
			if(entry->eob) { 							// eob => take out of list and delete
				if(gi->debugLevel > 1)
					fprintf(stderr, "INFO: EOF and EOB of %s => removing.\n", entry->name);

				if(lastEntry == NULL) {
					pthread_rwlock_wrlock(&(gi->rwlock));
					gi->audioList = entry->next;
					pthread_rwlock_unlock(&(gi->rwlock));
				}
				else {
					pthread_rwlock_wrlock(&(gi->rwlock));
					lastEntry->next = entry->next;
					pthread_rwlock_unlock(&(gi->rwlock));
				}

//				sleep(0);  // Give up CPU so that other processes (jprocess) may get some control.

				delEntry = entry;
				entry = entry->next;
				destroy_audio_entry(delEntry);
			}
			else {
				lastEntry = entry;
				entry = entry->next;
			}
		}

		// Now - get some data....
		entry = gi->audioList;
		while(entry) {
			fillEntryBuffers(entry);
			entry = entry->next;
		}
	}

	return 0;
}

/**** void fillEntryBuffers(audio_entry *entry) ****
*  A helpter thread to disk_thread.
*  It fill the buffers of a audio_entry.
*  Is put into this function, so the buffers can be
*  filled from start from open_audio_entry(...).
***************************************************/
void fillEntryBuffers(audio_entry *entry) {
	jack_ringbuffer_t **ringbuffer;
	jack_default_audio_sample_t *ptr, *ptr_eof;
	uint availSpace, framesRead, doCopyFrames;
	uint i, j;
	global_info *gi = entry->gi;

	// Check that we have not prev marked this as eof
	if(entry->eof)
		return;
			
	// Find minimum space in buffers
	ringbuffer = entry->ringbuffers;
	availSpace = jack_ringbuffer_write_space(*ringbuffer);
	ringbuffer++;
	for(i = 1; i < entry->channels; i++) {
		if((j = jack_ringbuffer_write_space(*ringbuffer)) < availSpace)
			availSpace = j;
		ringbuffer++;
	}

	while(availSpace > gi->audioSampleSize && entry->eof == 0) {
		if(availSpace < gi->internalBufferSize)
			doCopyFrames = availSpace / gi->audioSampleSize;
		else
			doCopyFrames = gi->internalBufferSize / gi->audioSampleSize;

		framesRead = sf_readf_float(entry->sf_sndfile, (float *) gi->internalInterlacedBuffer, doCopyFrames);
		if(framesRead != doCopyFrames) {
			if (entry->count == 0) {
				entry->eof = 1;
				if(gi->debugLevel > 1)
					fprintf(stderr, "INFO: EOF of %s.\n", entry->name);
			} else {
				entry->count--;
				sf_seek(entry->sf_sndfile, 0, SEEK_SET);
			}
		}
				
		ptr = (jack_default_audio_sample_t *) gi->internalInterlacedBuffer;
		ptr_eof = (jack_default_audio_sample_t *) (ptr + framesRead * entry->channels);
		j = 0;

		while(ptr < ptr_eof) {
			for(i = 0; i < entry->channels; i++) {
				*((jack_default_audio_sample_t *) (gi->internalBufferArray + i * gi->internalBufferSize + j * gi->audioSampleSize)) = *ptr * entry->volume;
				ptr++;
			}
			j++;
		}
				
		if(j != framesRead) 
			fprintf(stderr, "ERROR: Frames intented to read and frames actually read into internal buffer differs!\n"
			                "       Intended to read: %u. Actually read: %u.\n", framesRead, j);

		ringbuffer = entry->ringbuffers;
		j *= gi->audioSampleSize;
		for(i = 0; i < entry->channels; i++) {
			jack_ringbuffer_write(*ringbuffer, gi->internalBufferArray + i * gi->internalBufferSize, j);
			ringbuffer++;
		}

		availSpace -= j;
	}
}

/*** void destroy_audio_entry(audio_entry *entry) ****
*  As the name indicates...
******************************************************/
void destroy_audio_entry(audio_entry *entry) {
	jack_ringbuffer_t **ringbuffer;
	uint i = 0;

	if(entry->name)
		free(entry->name);

	if(entry->sf_sndfile)
		sf_close(entry->sf_sndfile);

	if(entry->sf_info)
		free(entry->sf_info);

	if(entry->ringbuffers) {
		ringbuffer = entry->ringbuffers;
		for(i = 0 ; i < entry->channels; i++) {
			if(*ringbuffer)
				jack_ringbuffer_free(*ringbuffer);
			ringbuffer++;
		}
		free(entry->ringbuffers);
	}

	if((entry->vu))
		free(entry->vu);


	if(entry->portConnections) {
		i = 0;
		while((entry->portConnections + i)->last == 0) {
			if((entry->portConnections + i)->dstName) {
				free((entry->portConnections + i)->dstName);
				(entry->portConnections + i)->dstName = 0;
			}
			i++;
		}

		if((entry->portConnections + i)->dstName)
			free((entry->portConnections + i)->dstName);
			
		free(entry->portConnections);
	}


	free(entry);
}


/*********** void take_out_audio_entry(...) ************
*  Take out entry from the linked list and destroy it.
*  NB: Not any more - just mark it for deletion!
*******************************************************/
void take_out_audio_entry(audio_entry *delEntry) {
	delEntry->fillZero = 1;
}

/****** int open_audio_entry(...) ******
*  As the name indicates...
*  Should also allocate ringbuffers, jackports
*  and insert it to the chain.
*  BTW - it should also connect its output port to 
*  a jack input port.
*******************************************************/
audio_entry * open_audio_entry(global_info *gi, const char *filename, const char *name, port_connection *connections, float volume, signed int repeat_count) {
	uint i;
	audio_entry *entry;

	SF_INFO *sfinfo;
	SNDFILE *sndfile;
	jack_ringbuffer_t **ringbufferPtr;

	uint ringbufSize;

	entry = (audio_entry *) malloc(sizeof(audio_entry));
	if(entry == NULL)
		return NULL;

	entry->gi = gi;
	entry->name = NULL;
	entry->sf_info = NULL;
	entry->sf_sndfile = NULL;
	entry->ringbuffers = NULL;
	entry->portConnections = NULL;
	entry->next = 0;
	entry->currentFrame = 0;
	entry->error = 0;
	entry->count = repeat_count>0?repeat_count-1:0;
	entry->eof = 0;
	entry->eob = 0;
	entry->fillZero = 0;
	entry->volume = volume;
	entry->vu = 0;
	entry->gui_ptr = 0;

	if(gi->client == NULL)
		entry->error |= ERR_NO_JACK_CLIENT;

	if(filename == NULL)
		entry->error |= ERR_MISSING_FILENAME;

	if(entry->error)
		return entry;

	if(access(filename, F_OK)) {
		entry->error = ERR_FILE_NOT_FOUND;
		return entry;
	}

	if(name) {
		entry->name = (char *) malloc(strlen(name) + 1);
		if(entry->name == NULL) {
			entry->error |= ERR_COULD_NOT_ALLOCATE_MEM;
			return entry;
		}
		strcpy(entry->name, name);
	}

	sfinfo = (SF_INFO*) malloc(sizeof(SF_INFO));
	if(sfinfo == NULL) {
		entry->error |= ERR_COULD_NOT_ALLOCATE_MEM;
		return entry;
	}

	sfinfo->format = 0;

	sndfile = sf_open(filename, SFM_READ, sfinfo);
	if(sndfile) {
		if((uint) sfinfo->channels > gi->maxAudioChannels) {
			entry->error = ERR_TO_MANY_CHANNELS;
			free(sfinfo);
			sf_close(sndfile);
			return entry;
		}

		entry->vu = (jack_default_audio_sample_t *) malloc(gi->audioSampleSize * sfinfo->channels);
		if(entry->vu == NULL)
			entry->error |= ERR_COULD_NOT_ALLOCATE_MEM;

		for(i = 0; i < (uint) sfinfo->channels; i++)
			*(entry->vu + i) = 0.0;

		if(sf_command(sndfile, SFC_SET_NORM_FLOAT, NULL, SF_TRUE) != 1) {
			entry->error = ERR_SETTING_NORMALIZE;
			free(sfinfo);
			free(entry->vu);
			sf_close(sndfile);
			return entry;
		}


		// Account for the repeat count in the total frames count
		entry->totalFrames = sfinfo->frames * (entry->count+1);
		entry->samplerate = sfinfo->samplerate;
		entry->channels = sfinfo->channels;
		entry->sf_info = sfinfo;
		entry->sf_sndfile = sndfile;
		entry->id = gi->audioIdCounter++;


		// --------------------------------
		// Copy port connections
		// ------------------
		if(connections) {
			i = 0;
			while((connections + i)->last == 0)
				i++;
			i++;

			entry->portConnections = (port_connection *) malloc(i * sizeof(port_connection));
			if(entry->portConnections) {
				i = 0;
				while((connections + i)->last == 0) {
					(entry->portConnections + i)->srcChannel = (connections + i)->srcChannel;
					(entry->portConnections + i)->last = 0;
					if((connections + i)->dstName) {
						(entry->portConnections + i)->dstName = (char *) malloc(strlen((connections + i)->dstName) + 1);
						strcpy((entry->portConnections + i)->dstName, (connections + i)->dstName);
					}
					else
						(entry->portConnections + i)->dstName = "-err-";

					i++;
				}
				(entry->portConnections + i)->srcChannel = (connections + i)->srcChannel;
				(entry->portConnections + i)->last = 1;
				if((connections + i)->dstName) {
					(entry->portConnections + i)->dstName = (char *) malloc(strlen((connections + i)->dstName) + 1);
					strcpy((entry->portConnections + i)->dstName, (connections + i)->dstName);
				}
				else
					(entry->portConnections + i)->dstName = "-err-";
			}
			else
				fprintf(stderr, "Could not allocate memory for port info!\n");
		}
		else if(gi->debugLevel > 1)
			fprintf(stderr, "Warning: No ports assigned!\n");



		// --------------------------------
		// Create ringbuffers
		// ------------------

		// Allocate space for ringbuffers and point to NULL
		entry->ringbuffers = malloc(sizeof(jack_ringbuffer_t *) * entry->channels);
		if(entry->ringbuffers == NULL) {
			entry->error |= ERR_COULD_NOT_ALLOCATE_MEM;
			return entry;
		}

		ringbufferPtr = entry->ringbuffers;
		for(i = 0; i < entry->channels; i++) {
			*ringbufferPtr = NULL;
			ringbufferPtr++;
		}

		ringbufferPtr = entry->ringbuffers;
		ringbufSize = gi->ringbufferSizeFactor * gi->jackBufSize * gi->audioSampleSize;
		for(i = 0; i < entry->channels; i++) {
			if(gi->debugLevel > 1)
				fprintf(stderr, "INFO: Requesting ringbuffer (size: %u)... ", ringbufSize);
			
			if((*ringbufferPtr = jack_ringbuffer_create(ringbufSize))) {
				if(gi->debugLevel > 1)
					fprintf(stderr, "OK\n");
			}
			else {
				if(gi->debugLevel > 1)
					fprintf(stderr, "FAILED!\n");

				entry->error |= ERR_COULD_NOT_RINGBUF;
				return entry;
			}
			ringbufferPtr++;
		}

		
		// ----------------------------------
		// Fill buffers and insert into chain
		// ------------------
		fillEntryBuffers(entry);
		pthread_rwlock_rdlock(&(gi->rwlock));
		entry->next = gi->audioList;
		gi->audioList = entry;
		pthread_rwlock_unlock(&(gi->rwlock));
	}
	else {
		entry->error = ERR_FILE_COULD_NOT_SFOPEN;
		free(sfinfo);
	}

	return entry;
}

/*********** file_info *get_file_info(...) ***********
*  Get fileinfo.
*  The pointer *filename will be used in 
*  file_info->filename.
*****************************************************/
file_info *get_file_info(global_info *gi, const char *filename) {
	file_info *a;
	SF_INFO *sfinfo;
	SNDFILE *sndfile;

	a = (file_info*) malloc(sizeof(file_info));
	a->error = 0;

	if(filename == NULL) {
		a->error |= ERR_MISSING_FILENAME;
		return a;
	}

	if(access(filename, F_OK)) {
		a->error = ERR_FILE_NOT_FOUND;
		return a;
	}

	sfinfo = (SF_INFO*) malloc(sizeof(SF_INFO));
	sfinfo->format = 0;

	sndfile = sf_open(filename, SFM_READ, sfinfo);

	if(sndfile) {
		a->totalFrames = sfinfo->frames;
		a->samplerate = sfinfo->samplerate;
		a->channels = sfinfo->channels;
		if(a->channels > gi->maxAudioChannels)
			a->error = ERR_TO_MANY_CHANNELS;

		free(sfinfo);
		sf_close(sndfile);
	}
	else {
		a->error = ERR_FILE_COULD_NOT_SFOPEN;
		free(sfinfo);
	}

	return a;
}

/***** void get_rdlock() - void release_rdlock() *******
* Just convenience functions to obtain and release
* read lock.
********************************************************/
int get_rdlock(global_info *gi) {
	if(gi->thread_status &THREAD_STATUS_RWLOCK) {
		pthread_rwlock_rdlock(&(gi->rwlock));
		return 1;
	}

	return 0;
}

void release_rdlock(global_info *gi) {
	if(gi->thread_status &THREAD_STATUS_RWLOCK)
		pthread_rwlock_unlock(&(gi->rwlock));
}
