#include <qapplication.h>
#include <qmessagebox.h>
#include <qmenubar.h>

#include "jackctrl.h"
#include "x11routines.h"
#include "guiconfig.h"
#include "mwindow.h"
#include "playlistentry.h"

extern global_info gi;
extern GUIconfig *guicfg;

MainWindow::MainWindow() : QWidget(0, 0, WType_TopLevel),
	keyGrabber(0), nullTimer(new QTimer(this)),
	activeWindow(new ActiveWindow(this)),
	dialogFile(new QFileDialog(this, 0, true)) {
	qApp->setMainWidget(this);
 	connect(qApp, SIGNAL(aboutToQuit()), this, SLOT(qappAboutToQuit()));

	
	scroll = new QScrollBar(Qt::Vertical, this);
	playlist = new Playlist(this, scroll);

	menu_file = new QPopupMenu(this);
	menu_file->insertItem("&New", this, SLOT(file_new()));
	menu_file->insertItem("&Open...", this, SLOT(file_open()));
	menu_file->insertItem("&Save", this, SLOT(file_save()));
	menu_file->insertItem("Save&As...", this, SLOT(file_saveAs()));
	menu_file->insertSeparator();
	saveOnCloseId = 1;
	menu_file->insertItem("Save on &close", this, SLOT(file_toggleSaveOnExit()), 0, saveOnCloseId);
	menu_file->insertSeparator();
	menu_file->insertItem("&Exit", this, SLOT(close()));

	menu_edit = new QPopupMenu(this);
	menu_edit->insertItem("Add &entry...", playlist, SLOT(getFilename_s()));
	menu_edit->insertItem("Add &separator...", playlist, SLOT(addSep()));
	menu_edit->insertItem("&Remove all...", playlist, SLOT(removeAll()));
	menu_edit->insertSeparator();
	menu_edit->insertItem("&Kill all current playing...", this, SLOT(edit_killall()));
	menu_edit->insertSeparator();
	menu_edit->insertItem("&Preferences", guicfg, SLOT(show()));

	menu_misc = new QPopupMenu(this);
	menu_misc->insertItem("&Disconnect jack", this, SLOT(misc_disconnect()), 0, 1);
	menu_misc->insertItem("&Connect to jack", this, SLOT(misc_connect()), 0, 2);
	menu_misc->insertSeparator();
	menu_misc->insertItem("&Show Active Window", this, SLOT(misc_showActiveWindow()), 0, 3);
	connect(menu_misc, SIGNAL(aboutToShow()), this, SLOT(misc_updateContents()));

	menu_help = new QPopupMenu(this);
	menu_help->insertItem("&About", this, SLOT(help_about()));
	menu_help->insertItem("About &QT", this, SLOT(help_aboutqt()));

	menuBar = new QMenuBar(this);
	menuBar->insertItem("&File", menu_file);
	menuBar->insertItem("&Edit", menu_edit);
	menuBar->insertItem("&Misc", menu_misc);
	menuBar->insertItem("&Help", menu_help);

	menuBar->setFocusPolicy(QWidget::NoFocus);
	menu_file->setFocusPolicy(QWidget::NoFocus);
	menu_edit->setFocusPolicy(QWidget::NoFocus);
	menu_misc->setFocusPolicy(QWidget::NoFocus);
	menu_help->setFocusPolicy(QWidget::NoFocus);

	connect(nullTimer, SIGNAL(timeout()), this, SLOT(nullTimerTimeout()));
	nullTimer->start(0, true);

	updateCaption();
}

MainWindow::~MainWindow() {
	delete menuBar;
	delete menu_edit;
	delete menu_help;
	delete menu_misc;
	delete scroll;
	delete playlist;
	delete keyGrabber;
	delete nullTimer;
	delete activeWindow;
	delete dialogFile;
	delete menu_file;
}

int MainWindow::getAvailHeight() {
	return height() - menuBar->height();
}

void MainWindow::resizeEvent(QResizeEvent *e) {
	int h = e->size().height();
	int w = e->size().width();
	int menuh = menuBar->sizeHint().height();
	int scrollw = scroll->sizeHint().width();
	int availh = h - menuh;
	int availw = w - scrollw;

	scroll->resize(scrollw, availh);
	scroll->move(availw, menuh);

	playlist->resize(availw, availh);
	playlist->move(0, menuh);

	adjustWidth();
}

void MainWindow::updateCaption() {
    setCaption("TTtrigger - " + guicfg->file.name());
}

int MainWindow::run() {
	guicfg->setParent(this, playlist);
	guicfg->load();
	menu_file->setItemChecked(saveOnCloseId, guicfg->saveOnClose);

	resize(400, 400);
	show();
	activeWindow->show();

	int w = playlist->sizeHint().width() + scroll->sizeHint().width();
	if(w < 400)
		w = 400;
	resize(w, 400);

	playlist->updateGraphics();

	X11Rt::setClassName(this, QString("MainWindow"), QString("TTtrigger"));

	return qApp->exec();
}

void MainWindow::file_save() {
	guicfg->save();
}

void MainWindow::file_saveAs() {
	dialogFile->setSelection(guicfg->fileName);
	dialogFile->setFilter("Projects (*.ttt)");
	dialogFile->setCaption("Save as...");
	dialogFile->setMode(QFileDialog::AnyFile);

	if(dialogFile->exec() & QDialog::Accepted) {
		QString tmpStr = dialogFile->selectedFile();
		if(tmpStr.isEmpty())
			return;
	
		QFileInfo fi(tmpStr);
		if(fi.isDir())
			fi = QFileInfo(fi.absFilePath() + "/" + DEF_FILE_NAME);

		if(fi.exists()) {
			int ans = QMessageBox::question(this, "Overwrite?", "File " + fi.fileName() + " exists. Do you want to overwrite?", QMessageBox::Ok, QMessageBox::Cancel | QMessageBox::Default);

			if(ans == QMessageBox::Ok) {
				guicfg->fileName = fi.absFilePath();
				guicfg->file.setName(guicfg->fileName);
				guicfg->save();
				updateCaption();
			}
		}
		else {
			guicfg->fileName = fi.absFilePath();
			guicfg->file.setName(guicfg->fileName);
			guicfg->save();
			updateCaption();
		}
	}
}

void MainWindow::file_open() {
	QString msg = QString("Are you sure you want to close current project?\n");
	if(guicfg->saveOnClose)
		msg += "Your project will be saved before opening existing project.";
	else
		msg += "Your project will NOT be saved.\n(See File->SaveOnClose)";

	int ans = QMessageBox::question(this, "Close current project?", msg, QMessageBox::Ok, QMessageBox::Cancel | QMessageBox::Default);

	if(ans == QMessageBox::Ok) {
		if(guicfg->saveOnClose)
			guicfg->save(false);

		dialogFile->setFilter("Projects (*.ttt)");
		dialogFile->setCaption("Open project...");
		dialogFile->setMode(QFileDialog::ExistingFile);

		if(dialogFile->exec() & QDialog::Accepted) {
			QString tmpStr = dialogFile->selectedFile();
			if(tmpStr.isEmpty())
				return;

			playlist->removeAll(false);
			guicfg->setDefaults();
			menu_file->setItemChecked(saveOnCloseId, guicfg->saveOnClose);
			guicfg->fileName = tmpStr;
			guicfg->file.setName(tmpStr);
			guicfg->load();
			menu_file->setItemChecked(saveOnCloseId, guicfg->saveOnClose);
			playlist->updateGraphics();
			updateCaption();
		}
	}
}

void MainWindow::file_new() {
	QString msg = QString("Are you sure you want to close current project?\n");
	if(guicfg->saveOnClose)
		msg += "Your project will be saved..";
	else
		msg += "Your project will NOT be saved.\n(See File->SaveOnClose)";

	int ans = QMessageBox::question(this, "Close current project?", msg, QMessageBox::Ok, QMessageBox::Cancel | QMessageBox::Default);

	if(ans == QMessageBox::Ok) {
		if(guicfg->saveOnClose)
			guicfg->save(false);

		dialogFile->setFilter("Projects (*.ttt)");
		dialogFile->setCaption("Set new project file...");
		dialogFile->setMode(QFileDialog::AnyFile);

		if(dialogFile->exec() & QDialog::Accepted) {
			QString tmpStr = dialogFile->selectedFile();
			if(tmpStr.isEmpty())
				return;

			QFileInfo fi(tmpStr);
			if(fi.isDir())
				fi = QFileInfo(fi.absFilePath() + "/" + DEF_FILE_NAME);

			if(fi.exists()) {
				int ans = QMessageBox::question(this, "Overwrite?", "File " + fi.fileName() + " exists. Overwrite file?", QMessageBox::Ok, QMessageBox::Cancel | QMessageBox::Default);

				if(ans == QMessageBox::Ok) {
					playlist->removeAll(false);
					guicfg->setDefaults();
					menu_file->setItemChecked(saveOnCloseId, guicfg->saveOnClose);
					guicfg->fileName = fi.absFilePath();
					guicfg->file.setName(guicfg->fileName);
					playlist->updateGraphics();
					updateCaption();
				}
			}
			else {
				playlist->removeAll(false);
				guicfg->setDefaults();
				menu_file->setItemChecked(saveOnCloseId, guicfg->saveOnClose);
				guicfg->fileName = fi.absFilePath();
				guicfg->file.setName(guicfg->fileName);
				playlist->updateGraphics();
				updateCaption();
			}
		}
	}
}

void MainWindow::file_toggleSaveOnExit() {
	guicfg->saveOnClose = !guicfg->saveOnClose;
	menu_file->setItemChecked(saveOnCloseId, guicfg->saveOnClose);
}

void MainWindow::edit_killall() {
	int ans = QMessageBox::question(this, "Kill all?", "Are you sure you want to kill ALL?", QMessageBox::Ok, QMessageBox::Cancel | QMessageBox::Default);
	if(ans == QMessageBox::Ok)
		activeWindow->aPlaylist->killAll();
}

void MainWindow::help_about() {
	QMessageBox::about( this, "TTtrigger by JVC. (v" APPVERSION ")",
		"TTtrigger v" APPVERSION ":\n"
	    "\n"
		"Written Summer 2005.\n"
		"\n"
		"Best regards\n"
		"Jrn Christensen\n");
}

void MainWindow::help_aboutqt() {
	QMessageBox::aboutQt(this, "TTtrigger" );
}

void MainWindow::closeEvent(QCloseEvent *e) {
	QString msg = QString("Are you sure you want to quit?\n");
	if(guicfg->saveOnClose)
		msg += "Your project will be saved when exiting.";
	else
		msg += "Your project will NOT be saved when exiting.\n(See File->SaveOnClose)";

	int ans = QMessageBox::question(this, "Quit?", msg, QMessageBox::Ok, QMessageBox::Cancel | QMessageBox::Default);

	if(ans == QMessageBox::Ok)
		e->accept();
	else
		e->ignore();
}

void MainWindow::qappAboutToQuit() {
	if(guicfg->saveOnClose)
		guicfg->save(false);

	stop_jackctrl(&gi);
}

void MainWindow::misc_disconnect() {
	if(gi.client)
		stop_jackctrl(&gi);
}

void MainWindow::misc_connect() {
	if(gi.client == 0)
		if(init_jackctrl(&gi))
			QMessageBox::critical(this, "Error!", "Could not connect to jack! Please see stderr for more info.", QMessageBox::Ok, QMessageBox::NoButton);
}

void MainWindow::misc_showActiveWindow() {
	if(activeWindow->isVisible())
		activeWindow->hide();
	else
		activeWindow->show();
}

void MainWindow::misc_updateContents() {
	if(gi.client) {
		menu_misc->setItemEnabled(1, true);
		menu_misc->setItemEnabled(2, false);
	}
	else {
		menu_misc->setItemEnabled(1, false);
		menu_misc->setItemEnabled(2, true);
	}

	menu_misc->setItemChecked(3, activeWindow->isVisible());
}

void MainWindow::keyPressEvent(QKeyEvent * e) {
	if(e->key() == guicfg->key_ctrl) {
		playlist->keyModPressed = true;
		playlist->updateButtonGraphics();
		return;
	}
	else if(keyGrabber) {
		keyGrabber->pasteKey(e->key());
		return;
	}

	if (e->key() == guicfg->key_alloff) {
		activeWindow->aPlaylist->killAll();
	}

	if(e->key() == guicfg->key_play_next) {
		if(playlist->ptrCurrent && !playlist->keyModPressed) {
			playlist->ptrCurrent->button_play->animateClick();
			playlist->currentAdvance();
		}
	}

	if(e->key() == guicfg->key_skip) {
		if(playlist->keyModPressed)
			playlist->currentBack();
		else
			playlist->currentAdvance();

		e->accept();
	}

	if(playlist->keyModPressed) {
		if(e->key() == guicfg->key_kill1)
			activeWindow->aPlaylist->killNum(0);
		if(e->key() == guicfg->key_kill2)
			activeWindow->aPlaylist->killNum(1);
		if(e->key() == guicfg->key_kill3)
			activeWindow->aPlaylist->killNum(2);
		if(e->key() == guicfg->key_kill4)
			activeWindow->aPlaylist->killNum(3);
		if(e->key() == guicfg->key_kill5)
			activeWindow->aPlaylist->killNum(4);
		if(e->key() == guicfg->key_kill6)
			activeWindow->aPlaylist->killNum(5);
		if(e->key() == guicfg->key_kill7)
			activeWindow->aPlaylist->killNum(6);
		if(e->key() == guicfg->key_kill8)
			activeWindow->aPlaylist->killNum(7);
		if(e->key() == guicfg->key_kill9)
			activeWindow->aPlaylist->killNum(8);
		if(e->key() == guicfg->key_kill10)
			activeWindow->aPlaylist->killNum(9);
		if(e->key() == guicfg->key_kill11)
			activeWindow->aPlaylist->killNum(10);
		if(e->key() == guicfg->key_kill12)
			activeWindow->aPlaylist->killNum(11);
	}

	for(PlaylistBase *it = playlist->itemFirst; it; it = it->next)
		if(it->isEnabled())
			if(((PlaylistEntry *) it)->hotkey == e->key())
				((PlaylistEntry *) it)->button_play->animateClick();
}

void MainWindow::keyReleaseEvent (QKeyEvent * e) {
	if(e->key() == guicfg->key_ctrl) {
		playlist->keyModPressed = false;
		playlist->updateButtonGraphics();
		e->accept();
		return;
	}

	e->ignore();
}

void MainWindow::nullTimerTimeout() {
	if(gi.client == 0)
		QMessageBox::critical(this, "Error!", "Could not connect to JACK. Please see stderr for more info.", QMessageBox::Ok, QMessageBox::NoButton);
}

void MainWindow::adjustWidth() {
	int new_w = playlist->getMinWidth() + scroll->width();

	if(new_w > width())
		resize(new_w, height());
}
