#include <qapplication.h>
#include <qfileinfo.h>
#include <qmessagebox.h>
#include <qinputdialog.h>
#include <cstdlib>

#include "playlist.h"
#include "mwindow.h"
#include "jackctrl.h"
#include "playlistsep.h"

extern global_info gi;

Playlist::Playlist(QWidget *p, QScrollBar *scb) : QWidget(p), scroll(scb), 
	parent(p),
	isUpdatingGraphics(false), 
	ptrCurrent(0), ptrPrev(0), ptrNext(0),
	itemFirst(0), itemLast(0), itemCount(0),
	keyModPressed(false) {

	scroll->setEnabled(false);
	scroll->setMinValue(0);
	scroll->setValue(0);

	connect(scroll, SIGNAL(valueChanged(int)), this, SLOT(updateLayout()));

	portSelectMenu = new PortQuickSelect(this);
	portAdvanced = new AdvPortSelect(this);
	widgetJumpTo = new WidgetJumpTo(this);
	popupName = new PopupName(this);

	popupMenu = new PopupMenu(this);
}

Playlist::~Playlist() {
	delete portSelectMenu;
	delete portAdvanced;
	delete widgetJumpTo;
	delete popupName;
	delete popupMenu;
}

void Playlist::updateScroll() {
	int height = 0;
	uint num = 0;

	for(PlaylistBase *it = itemFirst; it; it = it->next) {
		height += it->sizeHint().height();
		if(height >= ((MainWindow *) parent)->getAvailHeight())
			break;
		num++;
	}

	if(num < itemCount) {
		scroll->setMaxValue(itemCount - num);
		scroll->setEnabled(true);
	}
	else {
		scroll->setMaxValue(0);
		scroll->setEnabled(false);
	}

	updateLayout();
}

void Playlist::updateLayout() {
	int offset = 0;
	PlaylistBase *it;

	if(scroll->isEnabled()) {
		int i = 0;
		int skip = scroll->value();

		it = itemFirst;
		while(i < skip && it != 0) {
			it->hide();
			it = it->next;
			i++;
		}

		while(offset < ((MainWindow *) parent)->getAvailHeight() && it != 0) {
			it->move(0, offset);
			it->show();
			offset += it->height();
			it = it->next;
		}

		
		while(it) {
			it->hide();
			it = it->next;
		}
	}
	else {
		for(it = itemFirst; it; it = it->next) {
			it->move(0, offset);
			it->show();
			offset += it->height();
		}
	}
}

void Playlist::addEntry(PlaylistBase *entry) {
	PlaylistBase *it;

	if(itemCount == 0) {
		itemFirst = entry;
		itemLast = entry;
		entry->next = 0;
		entry->prev = 0;
	}
	else if(entry->index == 0) {
		entry->next = itemFirst;
		if(itemFirst)
			itemFirst->prev = entry;
		itemFirst = entry;
		entry->prev = 0;
	}
	else if(entry->index < itemCount) {
		for(it = itemFirst; it; it = it->next)
			if(it->index == entry->index)
				break;

		entry->next = it;
		entry->prev = it->prev;
		entry->next->prev = entry;
		entry->prev->next = entry;
	}
	else {
		entry->prev = itemLast;
		if(itemLast)
			itemLast->next = entry;
		itemLast = entry;
		entry->next = 0;
	}
	itemCount++;		

	uint i = 0;
	for(it = itemFirst; it; it = it->next)
		it->index = i++;

	updateGraphics();
	updateScroll();

	if(sizeHint().width() > width())
		parent->resize(sizeHint().width() + ((MainWindow *) parent)->scroll->sizeHint().width(), height());
}

void Playlist::remEntry(PlaylistBase *entry) {
	if(QMessageBox::warning(this, "Delete entry?", "Do you want to remove " + entry->label_name->text() + "?", QMessageBox::Ok, QMessageBox::Cancel | QMessageBox::Escape | QMessageBox::Default) == QMessageBox::Cancel)
		return;

	entry->hide();
	remFromList(entry);

	if(entry->isEnabled()) {
		if(entry == ptrCurrent) {
			ptrCurrent = 0;
			currentAdvance();
		}
		else if(entry == ptrPrev)
			ptrPrev = 0;
		else if(entry == ptrNext)
			ptrNext = 0;
	}

	delete entry;

	updateGraphics();
	updateScroll();
}

void Playlist::remFromList(PlaylistBase *entry) {
	if(itemCount == 1) {
		itemFirst = 0;
		itemLast = 0;
	}
	else if(entry == itemFirst) {
		itemFirst = entry->next;
		itemFirst->prev = 0;
	}
	else if(entry == itemLast) {
		itemLast = entry->prev;
		itemLast->next = 0;
	}
	else {
		entry->prev->next = entry->next;
		entry->next->prev = entry->prev;
	}
	itemCount--;

	uint i = 0;
	for(PlaylistBase *it = itemFirst; it; it = it->next)
		it->index = i++;
}

void Playlist::removeAll(bool confirm) {
	if(confirm)
		if(QMessageBox::warning(this, "Delete all?", "Are you sure you want to remove ALL entries?", QMessageBox::Ok, QMessageBox::Cancel | QMessageBox::Escape | QMessageBox::Default) == QMessageBox::Cancel)
			return;

	PlaylistBase *n;
	for(PlaylistBase *it = itemFirst; it; it = n) {
		it->hide();
		n = it->next;
		delete it;
	}

	ptrPrev = 0;
	ptrCurrent = 0;
	ptrNext = 0;

	itemFirst = 0;
	itemLast = 0;

	itemCount = 0;

	updateGraphics();
	updateScroll();
}

void Playlist::openAddEntry(QString filename, QString name, ChannelPortList chnVSport, int h, int v, int r) {
	PlaylistEntry *n = new PlaylistEntry(this);
	n->filename = filename;
	n->listChnPort = chnVSport;
	n->label_name->setText(name);
	n->spinbox_repeatcount->setValue(r);
	n->hotkey = h;
	n->dial_vol->setValue(v);

	if(openEntry(n))
		addEntry(n);
	else
		delete n;
}

bool Playlist::openEntry(PlaylistEntry *entry) {
	qHeapSort(entry->listChnPort);

	file_info *file = get_file_info(&gi, entry->filename.ascii());

	if(entry->label_name->text().isEmpty()) {
		QFileInfo qfileinfo(entry->filename);
		entry->label_name->setText(qfileinfo.baseName());
	}

	entry->errorFlag = file->error;

	if(file->error) {
		if(file->error & ERR_MISSING_FILENAME) {
			if(gi.debugLevel > 0)
				qWarning("ERROR: Missing filename while opening audio file.");

			free(file);
			entry->totalFrames = 0;
			entry->samplerate = 0;
			entry->channels = 0;
			entry->setEnabled(false);
			return false;
		}

		if(file->error & ERR_FILE_NOT_FOUND) {
			if(gi.debugLevel > 0)
				qWarning("ERROR: File %s not found.", entry->filename.ascii());

			free(file);
			entry->totalFrames = 0;
			entry->samplerate = 0;
			entry->channels = 0;
			entry->setEnabled(false);
			return true;
		}

		if(file->error & ERR_FILE_COULD_NOT_SFOPEN) {
			if(gi.debugLevel > 0)
				qWarning("ERROR: Could not open audio file %s.", entry->filename.ascii());

			free(file);
			entry->totalFrames = 0;
			entry->samplerate = 0;
			entry->channels = 0;
			entry->setEnabled(false);
			return true;
		}

		if(file->error & ERR_TO_MANY_CHANNELS) {
			if(gi.debugLevel > 0)
				qWarning("WARNING: Cannot play file: %s\n"
						 "         To many channels!\n"
						 "         Increase number of allowed channels with -c.", entry->filename.ascii());
			entry->setEnabled(false);
		}
	}

	entry->totalFrames = file->totalFrames;
	entry->samplerate = file->samplerate;
	entry->channels = file->channels;
	entry->setEnabled(true);

	if(entry->samplerate != gi.jack_samplerate && gi.debugLevel > 1)
		qWarning("WARNING: Samplerate of file %s and jack does not match!", entry->filename.ascii());

	free(file);

	return true;
}

void Playlist::insertSep(QString name, PlaylistBase *entry) {
	PlaylistSep *n = new PlaylistSep(this, name);
	if(entry)
		n->index = entry->index;

	addEntry(n);
}

void Playlist::resizeEvent(QResizeEvent *) {
	updateScroll();
}

void Playlist::updateButtonGraphics() {
	for(PlaylistBase *it = itemFirst; it; it = it->next)
		it->updateButtonGraphics();
}

void Playlist::updateGraphics() {
	if(itemCount == 0 || isUpdatingGraphics)
		return;

	isUpdatingGraphics = true;

	if(ptrCurrent == 0 && ptrPrev == 0 && ptrNext == 0)
		setCurrentFirstEnabled();

	fixPrevNext();


	PlaylistBase *it;
	PEntryWidth wWidth;

	for(it = itemFirst; it; it = it->next) {
		if(it->isEntry) {
			((PlaylistEntry *) it)->updateWidgets();
			((PlaylistEntry *) it)->compareSize(&wWidth);
		}
		else {
			it->updateWidgets();
			it->compareSize(&wWidth);
		}
	}

	wWidth.width_center = wWidth.width_name + wWidth.width_filename + wWidth.width_repeatcount + wWidth.width_length + wWidth.width_play +
							wWidth.width_hotkey + wWidth.width_portcols;

	for(it = itemFirst; it; it = it->next) {
		if(it->isEntry)
			((PlaylistEntry *) it)->adjustSize(&wWidth);
		else
			it->adjustSize(&wWidth);

		it->layout->activate();
		it->resize(it->sizeHint());
	}

	((MainWindow *) parent)->adjustWidth();

	isUpdatingGraphics = false;
}

int Playlist::getMinWidth() {
	int w = 0, itw;
	for(PlaylistBase *it = itemFirst; it; it = it->next)
		if((itw = it->sizeHint().width()) > w)
			w = itw;

	return w;	
}

void Playlist::moveItem(PlaylistBase *item) {
	remFromList(item);
	addEntry(item);
}

void Playlist::getFilename_s(PlaylistBase *n, bool updateItem) {
	PlaylistEntry *entry;
	QFileDialog *dialog = ((MainWindow *)  qApp->mainWidget())->dialogFile;
	if(n != 0 && updateItem && n->isEntry) {
		entry = (PlaylistEntry *) n;

		dialog->setSelection(entry->filename);
		dialog->setFilter("Sounds (*.wav)");
		dialog->setCaption("Please select file...");
		dialog->setMode(QFileDialog::ExistingFile);

		if(dialog->exec() & QDialog::Accepted) {
			QString file = dialog->selectedFile();
			if(file.isEmpty())
				return;

			entry->filename = file;

			openEntry(entry);
			if(entry->errorFlag)
				QMessageBox::warning(this, "Warning", "There was an error while opening file " + file + ". See stderr for more information.", QMessageBox::Ok, QMessageBox::NoButton);

			if(!entry->isEnabled() && entry == ptrCurrent)
				currentAdvance();  // Does also updateGraphics()
			else
				updateGraphics();
		}

		return;
	}
	else {
		QStringList files, filesWithError;
		bool doAdd, doInsert;
		uint insertId = 0;

		if((doInsert = (n != 0)))
			insertId = n->index;

		dialog->setFilter("Sounds (*.wav)");
		dialog->setCaption("Please select one or more files...");
		dialog->setMode(QFileDialog::ExistingFiles);

		if(dialog->exec() & QDialog::Accepted) {
			files = dialog->selectedFiles();
			if(files.count() == 0)
				return;

			QStringList::iterator it = files.begin();
			while(it != files.end()) {
				entry = new PlaylistEntry(this);
				entry->filename = (QString) (*it);

				doAdd = openEntry(entry);
				if(entry->errorFlag)
					filesWithError << (*it);

				if(doAdd) {
					if(doInsert)
						entry->index = insertId++;
					addEntry(entry);
				}
				else
					delete entry;

				it++;
			}

			if(filesWithError.count()) {
				QString message = "There was an error while opening following files:\n" +
					filesWithError.join("\n") +
					"\n\nSee stderr for more information.";

				QMessageBox::warning(this, "Warning",  message, QMessageBox::Ok, QMessageBox::NoButton);
			}

			updateGraphics();
		}
	}
}

void Playlist::addSep(PlaylistBase *entry) {
	bool ok;
	QString newName = QInputDialog::getText("Enter title" , "Please enter title for separator", QLineEdit::Normal, "- Unnamed -", &ok, this);
	if(ok) {
		newName = newName.stripWhiteSpace();
		if(newName.isEmpty())
			insertSep("- Unnamed -", entry);
		else
			insertSep(newName, entry);
	}
}


void Playlist::popupNameEdit(PlaylistBase *current, bool next) {
	if(next) {
		current = current->next;
		if(current == 0)
			current = itemFirst;
	}
	else {
		current = current->prev;
		if(current == 0)
			current = itemLast;
	}
	popupName->popup(current);
}

void Playlist::setCurrent(PlaylistBase *entry) {
	if(entry->isEnabled()) {
		if(ptrCurrent != entry) {
			if(ptrCurrent)
				ptrCurrent->setCFlag(cBlank);

			ptrCurrent = (PlaylistEntry *) entry;

			if(entry == ptrPrev)
				ptrPrev = 0;

			if(entry == ptrNext)
				ptrNext = 0;

			ptrCurrent->setCFlag(cCurrent);
		}

		updateGraphics();
	}
}

void Playlist::fixPrevNext() {
	if(ptrCurrent == 0)
		return;

	if(ptrPrev)
		ptrPrev->setCFlag(cBlank);

	if(ptrNext)
		ptrNext->setCFlag(cBlank);

	if((ptrPrev = currentGetPrev(ptrCurrent)) != 0)
		ptrPrev->setCFlag(cPrev);

	if((ptrNext = currentGetNext(ptrCurrent)) != 0)
		ptrNext->setCFlag(cNext);
}

PlaylistEntry * Playlist::currentGetNext(PlaylistEntry * offset) {
	for(PlaylistBase *it = offset->next; it; it = it->next)
		if(it->isEnabled())
			return (PlaylistEntry *) it;

	return 0;
}

PlaylistEntry * Playlist::currentGetPrev(PlaylistEntry * offset) {
	for(PlaylistBase *it = offset->prev; it; it = it->prev)
		if(it->isEnabled())
			return (PlaylistEntry *) it;

	return 0;
}

void Playlist::currentAdvance() {
	PlaylistBase *ptr = ptrNext;

	if (ptr==0 && ptrCurrent!=0) 
		ptr = currentGetNext(ptrCurrent);

	if (ptr != 0) {
		uint num = ptr->index;
		// When scrolling to follow the current entry, provide
		// 4 lines of "before" context.
		num = (num<4) ? 0 : num - 4;
		scroll->setValue(num);

		setCurrent(ptr);
	}
}

void Playlist::currentBack() {
	PlaylistBase *ptr = ptrPrev;

	if (ptr==0 && ptrCurrent!=0)
		ptr = currentGetPrev(ptrCurrent);

	if (ptr != 0) {
		uint num = ptr->index;
		// When scrolling to follow the current entry, provide
		// 4 lines of "before" context.
		num = (num < 4) ? 0 : num - 4;
		scroll->setValue(num);

		setCurrent(ptr);
	}
}

void Playlist::setCurrentFirstEnabled() {
	for(PlaylistBase *it = itemFirst; it; it = it->next)
		if(it->isEnabled()) {
			setCurrent(it);
			break;
		}
}
