#include <qapplication.h>
#include <qpainter.h>
#include <qbrush.h>
#include <qfileinfo.h>

#include "images.h"
#include "jackctrl.h"
#include "guiconfig.h"
#include "portselector.h"
#include "playlistentry.h"

extern global_info gi;
extern GUIconfig *guicfg;
extern Images *appImages;

PlaylistEntry::PlaylistEntry(Playlist *p) : PlaylistBase(p),
	totalFrames(0), samplerate(0), channels(0), hotkey(0) {
	isEntry = true;   // is PlaylistEntry
	label_length = new QLabel(this);
	button_play = new QPushButton(this);
	label_hotkey = new QLabel(this);
	label_portcols = new QLabel(this);
	spinbox_repeatcount = new QSpinBox(1, 999, 1, this);
	label_filename = new QLabel(this);
	dial_vol = new QDial(0, 140, 20, DEF_AUDIO_VOLUME, this);

	button_play->setPixmap(*(appImages->pixmap_play_xpm));

	QFont hFont;
	hFont.setWeight(QFont::Bold);
	label_index->setFont(hFont);
	label_name->setFont(hFont);

	spinbox_repeatcount->setFocusPolicy(NoFocus);

	label_hotkey->setFrameStyle(QFrame::Panel | QFrame::Raised );
	label_hotkey->setLineWidth(0);

	layout->addStretch();
	layout->add(label_current);
	layout->add(label_index);
	layout->add(label_name);
	layout->add(label_filename);
	layout->add(label_length);
	layout->add(spinbox_repeatcount);
	layout->add(dial_vol);
	layout->add(button_play);
	layout->add(label_hotkey);
	layout->add(label_portcols);
	layout->add(button_moveUp);
	layout->add(button_moveDown);
	layout->addStretch();

	button_play->setFlat(true);
	button_moveDown->setFlat(true);
	button_moveUp->setFlat(true);

	dial_vol->setFixedSize(30, 30);
	dial_vol->setNotchesVisible(true);
	dial_vol->setLineStep(20);

	dial_vol->setFocusPolicy(QWidget::NoFocus);
	button_play->setFocusPolicy(QWidget::NoFocus);
	button_moveDown->setFocusPolicy(QWidget::NoFocus);
	button_moveUp->setFocusPolicy(QWidget::NoFocus);

	timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(timeout()));
	connect(button_play, SIGNAL(clicked()), this, SLOT(play()));
}

PlaylistEntry::~PlaylistEntry() {
	delete label_length;
	delete label_filename;
	delete spinbox_repeatcount;
	delete button_play;
	delete label_hotkey;
	delete label_portcols;
	delete dial_vol;
	delete timer;
}

void PlaylistEntry::updateWidgets() {
	PlaylistBase::updateWidgets();

	uint minutes, seconds;
	if(gi.jack_samplerate)
		seconds = totalFrames / gi.jack_samplerate;
	else
		seconds = 0;
	minutes = seconds / 60;
	seconds %= 60;

	if(seconds < 10)
		label_length->setText("(" + QString::number(minutes) + ":0" + QString::number(seconds) + ")");
	else
		label_length->setText("(" + QString::number(minutes) + ":" + QString::number(seconds) + ")");


	if(hotkey)
		label_hotkey->setText(key2string(hotkey));
	else
		label_hotkey->setText(" - none - ");


	label_portcols->setPixmap(getPortCols(this, &listChnPort));

	QFileInfo fi(filename);
	label_filename->setText("[" + fi.fileName() + "]");

	resize(sizeHint());
}

void PlaylistEntry::setEnabled(bool val) {
	button_play->setEnabled(val);
}

bool PlaylistEntry::isEnabled() {
	return button_play->isEnabled();
}

QString PlaylistEntry::saveText() {
	QString out("[" + filename + "]\n");

	if(!label_name->text().isEmpty())
		out += "Name = " + label_name->text() + "\n";

	if(hotkey)
		out +="Hotkey = " + QString::number(hotkey) + "\n";

	out +="Repeats = " + spinbox_repeatcount->cleanText() + "\n";

	out +="Volume = " + QString::number(dial_vol->value()) + "\n";

	ChannelPortList::iterator it = listChnPort.begin();
	while(it != listChnPort.end()) {
		out += ">" + QString::number((*it).channel) + " " + (*it).name + "\n";
		it++;
	}

	return out;
}

void PlaylistEntry::updateButtonGraphics() {
	PlaylistBase::updateButtonGraphics();

	if(parent->keyModPressed) {
		button_play->setPixmap(*(appImages->pixmap_kill_xpm));
	}
	else {
		button_play->setPixmap(*(appImages->pixmap_play_xpm));
	}
}

void PlaylistEntry::mousePressEvent(QMouseEvent *e) {
	PlaylistBase::mousePressEvent(e);
	if(e->isAccepted())
		return;

	QWidget *clickedOn = childAt(mapFromGlobal(e->globalPos()));

	if(clickedOn == label_portcols) {
		if(parent->keyModPressed && button_play->isEnabled()) {
			parent->portSelectMenu->popup(e->globalPos(), this);
			e->accept();
			return;
		}
	}
	else if(clickedOn == label_filename) {
		if(parent->keyModPressed) {
			parent->keyModPressed = false;
			parent->updateButtonGraphics();
			parent->getFilename_s(this);
			e->accept();
			return;
		}
	}
	else if(clickedOn == label_hotkey) {
		if(parent->keyModPressed) {
			label_hotkey->setLineWidth(2);
			((MainWindow *) qApp->mainWidget())->keyGrabber = this;
			timer->start(TIMER_KEY_DELAY, true);
			e->accept();
			return;
		}
	}

	e->ignore();
}

void PlaylistEntry::mouseDoubleClickEvent(QMouseEvent *e) {
	PlaylistBase::mouseDoubleClickEvent(e);
	if(e->isAccepted())
		return;

	QWidget *clickedOn = childAt(mapFromGlobal(e->globalPos()));

	if(clickedOn == label_filename) {
		if(e->button() & Qt::LeftButton) {
			parent->getFilename_s(this);
			e->accept();
			return;
		}
	}
	else if(clickedOn == label_hotkey) {
		if(e->button() & Qt::LeftButton) {
			label_hotkey->setLineWidth(2);
			((MainWindow *) qApp->mainWidget())->keyGrabber = this;
			timer->start(TIMER_KEY_DELAY, true);
			e->accept();
			return;
		}
	}

	e->ignore();
}

void PlaylistEntry::compareSize(PEntryWidth *width) {
	PlaylistBase::compareSize(width);

	if(width->width_name < label_name->sizeHint().width())
		width->width_name = label_name->sizeHint().width();

	if(width->width_length < label_length->sizeHint().width())
		width->width_length = label_length->sizeHint().width();

	if(width->width_filename < label_filename->sizeHint().width())
		width->width_filename = label_filename->sizeHint().width();

	if(width->width_repeatcount < spinbox_repeatcount->sizeHint().width())
		width->width_repeatcount = spinbox_repeatcount->sizeHint().width();

	if(width->width_hotkey < label_hotkey->sizeHint().width())
		width->width_hotkey = label_hotkey->sizeHint().width();

	if(width->width_portcols < label_portcols->sizeHint().width())
		width->width_portcols = label_portcols->sizeHint().width();

	// Just need the width - they are all the same size
	width->width_play = button_play->width() + dial_vol->width();
}

void PlaylistEntry::adjustSize(PEntryWidth *width) {
	PlaylistBase::adjustSize(width);

	label_name->setFixedWidth(width->width_name);
	label_length->setFixedWidth(width->width_length);
	label_filename->setFixedWidth(width->width_filename);
	spinbox_repeatcount->setFixedWidth(width->width_repeatcount);
	label_hotkey->setFixedWidth(width->width_hotkey);
	label_portcols->setFixedWidth(width->width_portcols);
}

void PlaylistEntry::setCFlag(CurrentFlags f) {
	switch(f) {
		case cCurrent:
			label_current->setPixmap(*(appImages->pixmap_position_current_xpm));
			return;
		case cPrev:
			label_current->setPixmap(*(appImages->pixmap_position_prev_xpm));
			return;
		case cNext:
			label_current->setPixmap(*(appImages->pixmap_position_next_xpm));
			return;
		default:
			label_current->setPixmap(*(appImages->pixmap_position_blank_xpm));
			return;
	}
}

void PlaylistEntry::pasteKey(int key) {
	timer->stop();
	((MainWindow *) (qApp->mainWidget()))->keyGrabber = 0;
	label_hotkey->setLineWidth(0);
	hotkey = key;
	parent->updateGraphics();
}

void PlaylistEntry::timeout() {
	pasteKey(0);
}

void PlaylistEntry::play() {
	if(parent->keyModPressed)
		((MainWindow *) (qApp->mainWidget()))->activeWindow->aPlaylist->stop(label_name->text());
	else
		((MainWindow *) (qApp->mainWidget()))->activeWindow->aPlaylist->openEntry(filename, label_name->text(), listChnPort, dial_vol->value(), spinbox_repeatcount->value());
}
